from __future__ import annotations
import asyncio
import os
import pathlib
from datetime import datetime
from ftplib import FTP
from typing import Any, Dict, List, Optional, Set, Union, Tuple
from typing_extensions import Self

import humanize
from aioftp import Client
from loguru import logger
from pysus.data.local import Data
from tqdm import tqdm

CACHEPATH = os.getenv(
    "PYSUS_CACHEPATH", os.path.join(str(pathlib.Path.home()), "pysus")
)

__cachepath__ = pathlib.Path(CACHEPATH)
__cachepath__.mkdir(exist_ok=True)


def to_list(ite: Any) -> list:
    """Parse any builtin data type into a list"""
    return (
        [ite] if type(ite) in [str, float, int, Directory, File] else list(ite)
    )


class File:
    """
    FTP File class. This class will contain methods for interacting with
    files inside DataSUS FTP server. The databases will be responsible for
    parsing the files found for each db into File classes, enabling the
    databases' files to share state and its reusability.

    Parameters
        path [str]: entire directory path where the file is located
                              inside the FTP server
        name [str]: basename of the file
        info [dict]: a dict containing the keys [size, type, modify], which
                     are present in every FTP server. In PySUS, this info
                     is extract using `line_file_parser` with FTP LIST.

    Methods
        download(local_dir): extract the file to local_dir
        async_download(local_dir): async extract the file to local_dir
    """

    name: str
    extension: str
    basename: str
    path: str
    # parent: Directory # TODO: This causes too much overhead
    __info__: dict

    def __init__(self, path: str, name: str, info: dict) -> None:
        name, extension = os.path.splitext(name)
        self.name = name
        self.extension = extension
        self.basename = self.name + self.extension
        self.path = (
            path + self.basename
            if path.endswith("/")
            else path + "/" + self.basename
        )
        ppath = self.path.replace(self.basename, "")
        self.parent_path = (
            ppath[:-1]
            if ppath.endswith("/")
            else ppath
        )
        self.__info__ = info

    def __str__(self) -> str:
        return str(self.basename)

    def __repr__(self) -> str:
        return str(self.basename)

    def __hash__(self):
        return hash(self.path)

    def __eq__(self, other):
        if isinstance(other, File):
            return self.path == other.path
        return False

    @property
    def info(self):
        """
        Parse File info to human format
        """
        info = {}
        info["size"] = humanize.naturalsize(self.__info__["size"])
        info["type"] = self.extension[1:].upper() + " file"
        info["modify"] = self.__info__["modify"].strftime("%Y-%m-%d %I:%M%p")
        return info

    def download(self, local_dir: str = CACHEPATH, _pbar=None) -> Data:
        _dir = pathlib.Path(local_dir)
        _dir.mkdir(exist_ok=True, parents=True)
        filepath = _dir / self.basename
        filesize = int(self.__info__["size"])

        if _pbar:
            _pbar.unit = "B"
            _pbar.unit_scale = True
            _pbar.reset(total=filesize)

        _parquet = filepath.with_suffix(".parquet")
        if _parquet.exists():
            if _pbar:
                _pbar.update(filesize - _pbar.n)
            return Data(str(_parquet), _pbar=_pbar)

        _dbf = filepath.with_suffix(".dbf")
        if _dbf.exists():
            if _pbar:
                _pbar.update(filesize - _pbar.n)
            return Data(str(_dbf), _pbar=_pbar)

        if filepath.exists():
            if _pbar:
                _pbar.update(filesize - _pbar.n)
            return Data(str(filepath), _pbar=_pbar)

        if _pbar:
            _pbar.set_description(f"{self.basename}")

        try:
            ftp = FTP("ftp.datasus.gov.br")
            ftp.login()
            output = open(f"{filepath}", "wb")

            def callback(data):
                output.write(data)
                if _pbar:
                    _pbar.update(len(data))

            ftp.retrbinary(
                f"RETR {self.path}",
                callback,
            )
        except Exception as exc:
            raise exc
        finally:
            ftp.close()
            output.close()

        if _pbar:
            _pbar.update(filesize - _pbar.n)
        return Data(str(filepath), _pbar=_pbar)

    async def async_download(self, local_dir: str = CACHEPATH) -> Data:
        # aioftp.Client.parse_list_line_custom
        def line_file_parser(file_line):
            line = file_line.decode("utf-8")
            info = {}
            if "<DIR>" in line:
                date, time, _, *name = str(line).strip().split()
                info["size"] = 0
                info["type"] = "dir"
                name = " ".join(name)
            else:
                date, time, size, name = str(line).strip().split()
                info["size"] = size
                info["type"] = "file"

            modify = datetime.strptime(
                " ".join([date, time]), "%m-%d-%y %I:%M%p"
            )
            info["modify"] = modify.strftime("%m/%d/%Y %I:%M%p")

            return name, info

        _dir = pathlib.Path(local_dir)
        _dir.mkdir(exist_ok=True, parents=True)
        filepath = _dir / self.basename

        output = (
            local_dir + str(self.basename)
            if local_dir.endswith("/")
            else local_dir + "/" + str(self.basename)
        )

        _parquet = filepath.with_suffix(".parquet")
        if _parquet.exists():
            return Data(str(_parquet))

        _dbf = filepath.with_suffix(".dbf")
        if _dbf.exists():
            return Data(str(_dbf))

        if filepath.exists():
            return Data(output)

        async with Client.context(
            host="ftp.datasus.gov.br",
            parse_list_line_custom=line_file_parser,
        ) as client:
            await client.login()
            await client.download(self.path, output, write_into=True)

        return Data(output)


CACHE: Dict = {}


class Directory:
    """
    FTP Directory class. The Directory does not load its content when called.
    Instead, it will cache all the parents Directories until root "/". To load
    the content, the attr content or the method load() should be called. When
    firstly instantiated, it will CWD into the path provided and store self and
    all parents in cache

    Parameters
        path [str]: entire directory path where the directory is located
                    inside the FTP server
    Attrs
        name [str]: Directory name
        path [str]: Directory path
        parent [Directory]: parent Directory
        loaded [bool]: True if content is loaded
        content [dict[str:[File, Directory]]]: A dictionary with name and File
            or Directory inside the Directory (e.g: "name": Directory("name"))
    """

    name: str
    path: str
    parent: Directory
    loaded: bool = False
    __content__: Dict = {}

    def __new__(cls, path: str, _is_root_child=False) -> Directory:
        ftp = FTP("ftp.datasus.gov.br")
        path = f"/{path}" if not str(path).startswith("/") else path
        path = path[:-1] if path.endswith("/") else path

        if not path:  # if root ("/")
            path = "/"
            try:
                directory = CACHE["/"]
            except KeyError:
                directory = object.__new__(cls)
                directory.parent = directory
                directory.name = "/"
                directory.path = "/"
                directory.loaded = False
                directory.__content__ = {}
                CACHE["/"] = directory
            return directory

        parent_path, name = path.rsplit("/", maxsplit=1)

        if _is_root_child:
            # WARNING: This parameter is for internal meanings, do not use
            directory = object.__new__(cls)
            directory.parent = CACHE["/"]
            directory.name = name
            CACHE[path] = directory
            return directory

        try:
            directory = CACHE[path]  # Recursive and cached instantiation
        except KeyError:
            try:
                ftp.connect()
                ftp.login()
                ftp.cwd(path)  # Checks if parent dir exists on DATASUS
            except Exception as exc:
                if "cannot find the path" in str(exc):
                    logger.error(f"Not a directory {path}")
                elif "access is denied" in str(exc).lower():
                    #  Forbidden access, exists in ftp but returns Forbidden
                    directory = object.__new__(cls)
                    directory.parent = Directory(parent_path)  # Recursive
                    directory.name = name
                    directory.loaded = False
                    directory.__content__ = {}
                    CACHE[path] = directory
                    return directory
                raise exc
            finally:
                ftp.close()

            directory = object.__new__(cls)
            # TODO: In next step, all the parent directories will be generated,
            # but it cwds into every parent, while its certain that they exist
            # in ftp server. The best approach should be to skip the cwds
            directory.parent = Directory(parent_path)  # Recursive
            directory.name = name
            directory.loaded = False
            directory.__content__ = {}
            CACHE[path] = directory
        return directory

    def __init__(self, path: str, _is_root_child=False) -> None:
        path = f"/{path}" if not str(path).startswith("/") else path
        path = path[:-1] if path.endswith("/") else path
        if not path:
            path = "/"
        self.path = path

    def __str__(self) -> str:
        return self.path

    def __repr__(self) -> str:
        return self.path

    def __hash__(self):
        return hash(self.path)

    def __eq__(self, other):
        if isinstance(other, Directory):
            return self.path == other.path
        return False

    def __truediv__(self, path: str):
        if isinstance(path, str):
            path = f"/{path}" if not path.startswith("/") else path
            path = path[:-1] if path.endswith("/") else path
            return Directory(self.path + path)
        raise ValueError("Unsupported division")

    @property
    def content(self):
        """
        Returns a list of Files and Directories in the Directory, will load
        if needed
        """
        if not self.loaded:
            self.load()
        return list(self.__content__.values())

    def load(self):
        """
        The content of a Directory must be explicitly loaded
        """
        self.__content__ |= load_path(self.path)
        self.loaded = True
        return self

    def reload(self):
        """
        Reloads the content of the Directory
        """
        self.loaded = False
        return self.load()

    def is_parent(self, other: Union[Self, File]) -> bool:
        """
        Checks if Directory or File is inside (or at any subdir) of self.  
        """
        if self.path == "/":
            return True

        target = other
        while target.path != "/":

            if self.path == target.path:
                return True

            if isinstance(other, File):
                # TODO: Implement parent logic on File (too much overhead)
                target = Directory(other.parent_path)
            else:
                target = target.parent

        return False


CACHE["/"] = Directory("/")


def load_path(path: str) -> Dict[str, Union[Directory, File]]:
    """
    This method is responsible for listing all the FTP directory's.
    Converts the items found within a valid DATASUS path into `File`s or
    Directories, returning its content.
    """
    path = str(path)
    content = {}
    ftp = FTP("ftp.datasus.gov.br")

    try:
        ftp.connect()
        ftp.login()
        ftp.cwd(path)

        def line_file_parser(file_line):
            info = {}
            if "<DIR>" in file_line:
                date, time, _, *name = str(file_line).strip().split()
                info["size"] = 0
                info["type"] = "dir"
                name = " ".join(name)
                modify = datetime.strptime(
                    " ".join([date, time]), "%m-%d-%y %I:%M%p"
                )
                info["modify"] = modify
                xpath = (
                    path + name if path.endswith("/") else path + "/" + name
                )
                content[name] = Directory(xpath)
            else:
                date, time, size, name = str(file_line).strip().split()
                info["size"] = size
                info["type"] = "file"
                modify = datetime.strptime(
                    " ".join([date, time]), "%m-%d-%y %I:%M%p"
                )
                info["modify"] = modify
                content[name] = File(path, name, info)

        ftp.retrlines("LIST", line_file_parser)
    except Exception as exc:
        raise exc
    finally:
        ftp.close()

    upper_names = [n.upper() for n in content]
    to_remove = []
    for name in content:
        if ".DBF" in name.upper():
            if name.upper().replace(".DBF", ".DBC") in upper_names:
                to_remove.append(name)

    for name in to_remove:
        del content[name]

    return content


class Database:
    """
    Base class for PySUS databases. Contains common functions
    for accessing DataSUS FTP server. With this class, it is
    possible to construct database classes for different DataSUS
    files, sharing state and functionalities.

    Parameters
        ftp [FTP]: ftplib.FTP object for connecting in DataSUS server.
        name [str]: database name
        paths [list[Directory]]: server paths where the files are located
        files [list[Files]]: list of parsed Files from Database content
        metadata [dict]: dict containing database's metadata information

    Methods
        load(): Loads the database paths content to its own content
        describe(file): describes a file according to each database's
                        spec. Returns a dict with file information
        format(file): extracts from file name database related info, such as
                      year, month, UF and/or other useful info for the DB
        get_files(Any): filters files using database related format, depending
                        on the database's files specs
    """

    ftp: FTP
    name: str
    paths: Tuple[Directory, ...]
    metadata: dict
    __content__: Dict[str, Union[Directory, File]]

    def __init__(self) -> None:
        self.ftp = FTP("ftp.datasus.gov.br")
        self.__content__ = {}

    def __repr__(self) -> str:
        return f'{self.name} - {self.metadata["long_name"]}'

    @property
    def content(self) -> List[Union[Directory, File]]:
        """
        Lists Database content. The `paths` will be loaded if this property is
        called or if explicitly using `load()`. To add specific Directory inside
        content, `load()` the directory and call `content` again.
        """
        if not self.__content__:
            logger.info(
                "content is not loaded, use `load()` to load default paths"
            )
            return []
        return sorted(list(self.__content__.values()), key=str)

    @property
    def files(self) -> List[File]:
        """
        Lists Files inside content. To load a specific Directory inside
        content, just `load()` this directory and list files again.
        """
        return list(filter(lambda f: isinstance(f, File), self.content))

    def load(
        self, directories: Optional[Union[Directory, List[Directory]]] = None
    ) -> Database:
        """
        Loads specific directories to Database content. Will aggregate the
        files found within Directories into Database.content.
        """
        if not directories:
            directories = self.paths

        directories = to_list(directories)

        for i, path in enumerate(directories):
            if isinstance(path, str):
                path = Directory(path)
                directories[i] = path

            if not isinstance(path, Directory):
                raise ValueError("path must a valid DATASUS directory")

        for directory in directories:
            directory.load()
            self.__content__ |= directory.__content__
        return self

    def describe(self, file: File) -> dict:
        """
        Receives a `File` and returns a dict with its information,
        according to the database's specifications. This method is
        helpful to return the FTP's file in a humanized format

        Parameters
            file [File]: a `File` instance
        """
        ...

    def format(self, file: File) -> tuple:
        """
        Formats a File based on the database specifications,
        extracting its name's parameters given a pattern.

        Parameters
            file [File]: a `File` instance
        """
        ...

    def get_files(self, *args, **kwargs) -> list[File]:
        """
        Filters the list of `File`s according to each database file
        pattern, as UFs, Groups, Years, Months, etc. This method will
        also be responsible to look for wrong values within the file
        pattern and possible extra characters in its basename
        """
        ...

    def download(
        self, files: List[File], local_dir: str = CACHEPATH
    ) -> List[str]:
        """
        Downloads a list of Files.
        """
        files = to_list(files)
        pbar = tqdm(total=len(files), dynamic_ncols=True)
        dfiles = []
        for file in files:
            if isinstance(file, File):
                dfiles.append(file.download(local_dir=local_dir, _pbar=pbar))
        pbar.close()
        if len(dfiles) == 1:
            return dfiles[0]
        return dfiles

    async def async_download(
        self, files: List[File], local_dir: str = CACHEPATH
    ):
        """
        Asynchronously downloads a list of files
        """

        async def download_file(file):
            if isinstance(file, File):
                await file.async_download(local_dir=local_dir)

        tasks = [download_file(file) for file in files]
        await asyncio.gather(*tasks)



# -*- coding:utf-8 -*-
u"""
Created on 2023/12/12
by luabida
license: GPL V3 or Later
"""
import unittest
import datetime

from unittest.mock import patch, MagicMock

from pysus.ftp.databases.sinan import SINAN
from pysus.ftp import File


class TestSINANDatabase(unittest.TestCase):

    def test_sinan(self):
        mock_content = {
            "ACBIBR06.dbc": File(
                path="/dissemin/publicos/SINAN/DADOS/FINAIS/ACBIBR06.dbc",
                name="ACBIBR06.dbc",
                info={
                    'size': 28326,
                    'type': 'file',
                    'modify': datetime.datetime(2023, 1, 16, 14, 15)
                }
            ),
            "ACBIBR07.dbc": File(
                path="/dissemin/publicos/SINAN/DADOS/FINAIS/ACBIBR07.dbc",
                name="ACBIBR07.dbc",
                info={
                    'size': 673314,
                    'type': 'file',
                    'modify': datetime.datetime(2023, 1, 16, 14, 15)
                }
            ),
            "ACBIBR08.dbc": File(
                path="/dissemin/publicos/SINAN/DADOS/FINAIS/ACBIBR08.dbc",
                name="ACBIBR08.dbc",
                info={
                    'size': 1048406,
                    'type': 'file',
                    'modify': datetime.datetime(2023, 1, 16, 14, 15)
                }
            ),
        }

        with patch(
            'pysus.ftp.databases.sinan.SINAN',
            return_value=MagicMock(__content__=mock_content)
        ) as mock_sinan:
            sinan = SINAN()
            sinan.__content__ = mock_sinan().__content__

            descriptions = [sinan.describe(file) for file in sinan.files]
            expected_descriptions = [
                {'name': 'ACBIBR06.dbc',
                 'disease': 'Acidente de trabalho com material biológico',
                 'year': 2006,
                 'size': '28.3 kB',
                 'last_update': '2023-01-16 02:15PM'},
                {'name': 'ACBIBR07.dbc',
                 'disease': 'Acidente de trabalho com material biológico',
                 'year': 2007,
                 'size': '673.3 kB',
                 'last_update': '2023-01-16 02:15PM'},
                {'name': 'ACBIBR08.dbc',
                 'disease': 'Acidente de trabalho com material biológico',
                 'year': 2008,
                 'size': '1.0 MB',
                 'last_update': '2023-01-16 02:15PM'}
            ]

            self.assertEqual(descriptions, expected_descriptions)

            formats = [sinan.format(file) for file in sinan.files]
            expected_formats = [('ACBI', 2006), ('ACBI', 2007), ('ACBI', 2008)]
            self.assertEqual(formats, expected_formats)

            get_files = sinan.get_files(dis_code='ACBI', year=2006)
            self.assertEqual(get_files, [sinan.files[0]])



# -*- coding:utf-8 -*-
u"""
Created on 2023/12/12
by luabida
license: GPL V3 or Later
"""
import unittest
import datetime

from unittest.mock import patch, MagicMock

from pysus.ftp.databases.sinasc import SINASC
from pysus.ftp import File


class TestSINASCDatabase(unittest.TestCase):

    def test_sinasc(self):
        mock_content = {
            "DNAC1996.DBC": File(
                path="/dissemin/publicos/SINASC/NOV/DNRES/DNAC1996.DBC",
                name="DNAC1996.DBC",
                info={
                    'size': 247527,
                    'type': 'file',
                    'modify': datetime.datetime(2020, 1, 27, 12, 5)
                }
            ),
            "DNAC1997.DBC": File(
                path="/dissemin/publicos/SINASC/NOV/DNRES/DNAC1997.DBC",
                name="DNAC1997.DBC",
                info={
                    'size': 266815,
                    'type': 'file',
                    'modify': datetime.datetime(2020, 1, 27, 12, 5)
                }
            ),
            "DNAC1998.DBC": File(
                path="/dissemin/publicos/SINASC/NOV/DNRES/DNAC1998.DBC",
                name="DNAC1998.DBC",
                info={
                    'size': 242404,
                    'type': 'file',
                    'modify': datetime.datetime(2020, 1, 27, 12, 5)
                }
            ),
        }

        with patch(
            'pysus.ftp.databases.sinasc.SINASC',
            return_value=MagicMock(__content__=mock_content)
        ) as mock_sinasc:
            sinasc = SINASC()
            sinasc.__content__ = mock_sinasc().__content__

            descriptions = [sinasc.describe(file) for file in sinasc.files]
            expected_descriptions = [
                {'name': 'DNAC1996.DBC',
                 'group': 'Declarações de Nascidos Vivos',
                 'uf': 'Acre',
                 'year': 1996,
                 'size': '247.5 kB',
                 'last_update': '2020-01-27 12:05PM'},
                {'name': 'DNAC1997.DBC',
                 'group': 'Declarações de Nascidos Vivos',
                 'uf': 'Acre',
                 'year': 1997,
                 'size': '266.8 kB',
                 'last_update': '2020-01-27 12:05PM'},
                {'name': 'DNAC1998.DBC',
                 'group': 'Declarações de Nascidos Vivos',
                 'uf': 'Acre',
                 'year': 1998,
                 'size': '242.4 kB',
                 'last_update': '2020-01-27 12:05PM'}
            ]

            self.assertEqual(descriptions, expected_descriptions)

            formats = [sinasc.format(file) for file in sinasc.files]
            expected_formats = [
                ('DN', 'AC', 1996),
                ('DN', 'AC', 1997),
                ('DN', 'AC', 1998)
            ]
            self.assertEqual(formats, expected_formats)

            get_files = sinasc.get_files(group='DN', uf='AC', year=1996)
            self.assertEqual(get_files, [sinasc.files[0]])



import unittest
import pytest

import pandas as pd

from pysus.online_data.vaccine import download_covid


class VaccineTestCase(unittest.TestCase):
    pytest.mark.timeout(5)
    def test_Download(self):
        """Careful! this download can take a long time"""
        df = download_covid("BA", only_header=True)
        self.assertIsInstance(df, pd.DataFrame)
        self.assertEqual(df.shape, (10000, 42))


if __name__ == "__main__":
    unittest.main()



from sqlalchemy import DATE, INTEGER, NUMERIC, VARCHAR

# Variables extracted from docs/metadata/SINAN files with converted
# to sqlalchemy types
COLUMN_TYPE = {
    "ID_MUNICIP": INTEGER,
    "ID_UNIDADE": INTEGER,
    "DT_NOTIFIC": DATE,
    "NU_ANO": INTEGER,
    "DENGUE": INTEGER,
    "CS_ESCOLAR": INTEGER,
    "ANT_IDADE": NUMERIC(3),
    "ANT_RACA": VARCHAR(1),
    "ID_OCUPA_N": VARCHAR(6),
    "ESCOLMAE": VARCHAR(2),
    "ANT_PRE_NA": VARCHAR(1),
    "UF_PRE_NAT": VARCHAR(2),
    "MUN_PRE_NA": VARCHAR(6),
    "UNI_PRE_NA": VARCHAR(7),
    "ANTSIFIL_N": VARCHAR(1),
    "LAB_PARTO": VARCHAR(1),
    "LAB_TITU_2": NUMERIC(4),
    "LAB_DT3": DATE,
    "LAB_CONF": VARCHAR(1),
    "TRA_ESQUEM": VARCHAR(2),
    "TRA_DT": DATE,
    "TP_NOT": INTEGER,
    "SEM_NOT": INTEGER,
    "ID_AGRAVO": VARCHAR(6),
    "ANT_TRATAD": VARCHAR(1),
    "ANT_UF_CRI": VARCHAR(2),
    "ANT_MUNI_C": VARCHAR(6),
    "ANT_LOCAL_": NUMERIC((8, 0)),
    "LABC_SANGU": VARCHAR(1),
    "LABC_TIT_1": NUMERIC(4),
    "LABC_DT_1": DATE,
    "LABC_IGG": VARCHAR(1),
    "LABC_DT": DATE,
    "LABC_LIQUO": VARCHAR(1),
    "LABC_TIT_2": NUMERIC(4),
    "LABC_DT_2": DATE,
    "LABC_TITUL": VARCHAR(1),
    "LABC_EVIDE": VARCHAR(1),
    "LABC_LIQ_1": VARCHAR(1),
    "TRA_DIAG_T": VARCHAR(1),
    "CLI_ASSINT": VARCHAR(1),
    "CLI_ANEMIA": VARCHAR(1),
    "CLI_ESPLEN": VARCHAR(1),
    "CLI_OSTEO": VARCHAR(1),
    "CLI_RINITE": VARCHAR(1),
    "HEPATO": VARCHAR(1),
    "LESOES": VARCHAR(1),
    "CLI_OUTRO": VARCHAR(1),
    "SIN_OUTR_E": VARCHAR(20),
    "TRA_ESQU_1": VARCHAR(1),
    "DS_ESQUEMA": VARCHAR(30),
    "EVOLUCAO": VARCHAR(1),
    "EVO_DIAG_N": VARCHAR(1),
    "DT_OBITO": DATE,
    "DT_INVEST": DATE,
    "ANT_UF_1": VARCHAR(2),
    "MUN_1": VARCHAR(6),
    "ANT_UF_2": VARCHAR(2),
    "MUN_2": VARCHAR(6),
    "ANT_UF_3": VARCHAR(2),
    "MUN_3": VARCHAR(6),
    "PRESENCA": VARCHAR(1),
    "PARASITO": DATE,
    "HISTORIA": VARCHAR(1),
    "CONTROLE": VARCHAR(1),
    "MANIPULA": VARCHAR(1),
    "MAECHAGA": VARCHAR(1),
    "ORAL": VARCHAR(1),
    "ASSINTOMA": VARCHAR(1),
    "EDEMA": VARCHAR(1),
    "MENINGOE": VARCHAR(1),
    "POLIADENO": VARCHAR(1),
    "FEBRE": VARCHAR(1),
    "HEPATOME": VARCHAR(1),
    "SINAIS_ICC": VARCHAR(1),
    "ARRITMIAS": VARCHAR(1),
    "ASTENIA": VARCHAR(1),
    "ESPLENOM": VARCHAR(1),
    "CHAGOMA": VARCHAR(1),
    "OUTRO_SIN": VARCHAR(1),
    "OUTRO_ESP": VARCHAR(30),
    "DT_COL_DIR": DATE,
    "EXAME": VARCHAR(1),
    "MICRO_HEMA": VARCHAR(1),
    "OUTRO": VARCHAR(1),
    "DT_COL_IND": DATE,
    "XENODIAG": VARCHAR(1),
    "HEMOCULT": VARCHAR(1),
    "DT_COL_S1": DATE,
    "DT_COL_S2": DATE,
    "ELI_IGM_S1": VARCHAR(1),
    "ELI_IGG_S1": VARCHAR(1),
    "ELI_IGM_S2": VARCHAR(1),
    "ELI_IGG_S2": VARCHAR(1),
    "HEM_IGM_S1": VARCHAR(1),
    "HEM_IGG_S1": VARCHAR(1),
    "HEM_IGM_S2": VARCHAR(1),
    "HEM_IGG_S2": VARCHAR(1),
    "IMU_IGM_S1": VARCHAR(1),
    "TIT_IGM_S1": VARCHAR(5),
    "IMU_IGM_S2": VARCHAR(1),
    "TIT_IGM_S2": NUMERIC(5),
    "IMU_IGG_S1": VARCHAR(1),
    "TIT_IGG_S1": NUMERIC(5),
    "IMU_IGG_S2": VARCHAR(1),
    "TIT_IGG_S2": NUMERIC(5),
    "RESUL_HIS": DATE,
    "RES_HIST": VARCHAR(1),
    "ESPECIFICO": VARCHAR(1),
    "SINTOMATIC": VARCHAR(1),
    "DROGA": VARCHAR(1),
    "TEMPO": NUMERIC(3),
    "CON_TRIAT": VARCHAR(1),
    "BIOSSEG": VARCHAR(1),
    "FISCALIZA": VARCHAR(1),
    "MED_OUTRO": VARCHAR(1),
    "OUTRO_DES": VARCHAR(30),
    "CLASSI_FIN": VARCHAR(1),
    "CRITERIO": VARCHAR(2),
    "CON_PROVAV": VARCHAR(1),
    "CON_OUTRA": VARCHAR(30),
    "CON_LOCAL": VARCHAR(1),
    "TPAUTOCTO": VARCHAR(1),
    "COUFINF": VARCHAR(2),
    "COPAISINF": VARCHAR(4),
    "COMUNINF": VARCHAR(6),
    "CODISINF": VARCHAR(4),
    "CO_BAINFC": NUMERIC(8),
    "NOBAIINF": VARCHAR(60),
    "DOENCA_TRA": VARCHAR(1),
    "DT_ENCERRA": DATE,
    "DT_DIGITA": DATE,
    "DS_OBS": VARCHAR(255),
    "EPIZOOTIAS": VARCHAR(1),
    "ISOL_VIR_N": VARCHAR(1),
    "VETOR_A": VARCHAR(1),
    "VACINADO": VARCHAR(1),
    "DT_VACINA": DATE,
    "UF_VAC": VARCHAR(2),
    "MUNCI_VAC": VARCHAR(6),
    "UNID_VAC": NUMERIC((8, 0)),
    "DOR_ABDO_N": VARCHAR(1),
    "SINT_HEM_N": VARCHAR(1),
    "FAGET": VARCHAR(1),
    "EXCR_RENA_": VARCHAR(1),
    "HOSPITALIZ": VARCHAR(1),
    "DT_INTERNA": DATE,
    "UF": VARCHAR(2),
    "MUNICIPIO": VARCHAR(6),
    "HOSPITAL": VARCHAR(70),
    "BT": VARCHAR(6),
    "AST": NUMERIC(5),
    "BD": VARCHAR(6),
    "ALT": NUMERIC(5),
    "DT_COL_1": DATE,
    "S1_IGM": VARCHAR(1),
    "DT_COL_2": DATE,
    "S2_IGM": VARCHAR(1),
    "MAT_COLETA": VARCHAR(1),
    "DT_COLETA": DATE,
    "RES_ISOLAM": VARCHAR(1),
    "HISTOPA": VARCHAR(1),
    "IMUNOH": VARCHAR(1),
    "DT_PTPCR": DATE,
    "RES_PTPCR": VARCHAR(1),
    "CLASFIN_ES": VARCHAR(30),
    "LOCALIDADE": None,
    "CON_ATIVID": VARCHAR(1),
    "DTATEND": DATE,
    "NUATEND": NUMERIC((4, 0)),
    "DTSUSPEIC": DATE,
    "STHOSPITAL": VARCHAR(1),
    "DTINTERNA": DATE,
    "DTALTA": DATE,
    "UF_HOSP": VARCHAR(2),
    "MUN_HOSP": VARCHAR(6),
    "UNID_HOSP": NUMERIC(7),
    "STFEBRE": VARCHAR(1),
    "STNAUSEA": VARCHAR(1),
    "STVOMITO": VARCHAR(1),
    "STDIARREIA": VARCHAR(1),
    "STCONSTIPA": VARCHAR(1),
    "STCEFALEIA": VARCHAR(1),
    "STTONTURA": VARCHAR(1),
    "STVISAO": VARCHAR(1),
    "STDIPLOPIA": VARCHAR(1),
    "STDISARTRI": VARCHAR(1),
    "STDISFONIA": VARCHAR(1),
    "STDISFAGIA": VARCHAR(1),
    "STBOCA": VARCHAR(1),
    "STFERIMENT": VARCHAR(1),
    "STFLACIDEZ": VARCHAR(1),
    "STDISPNEIA": VARCHAR(1),
    "STRESPIRA": VARCHAR(1),
    "STCARDIACA": VARCHAR(1),
    "STCOMA": VARCHAR(1),
    "STPARESTES": VARCHAR(1),
    "DS_PARES": VARCHAR(30),
    "STOUTROSIN": VARCHAR(1),
    "DS_OUTROSI": VARCHAR(30),
    "STPTOSE": VARCHAR(1),
    "STOFTALMO": VARCHAR(1),
    "STMIDRIASE": VARCHAR(1),
    "STFACIAL": VARCHAR(1),
    "STBULBAR": VARCHAR(1),
    "STMEMSUP": VARCHAR(1),
    "STMEMINF": VARCHAR(1),
    "STDESCENDE": VARCHAR(1),
    "STSIMETRIC": VARCHAR(1),
    "STSENSIVEL": VARCHAR(1),
    "TPNEURO": VARCHAR(1),
    "STALIMENTO": VARCHAR(1),
    "DSALIMENTO": VARCHAR(30),
    "STCOMERCIO": VARCHAR(1),
    "STCASEIRA": VARCHAR(1),
    "DS_INDUS": VARCHAR(30),
    "STEXPALIM": VARCHAR(1),
    "DS_INGEST": VARCHAR(4),
    "DS_INI_GES": VARCHAR(4),
    "DS_FIM_GES": VARCHAR(4),
    "STDOMICILI": VARCHAR(1),
    "STESCOLA": VARCHAR(1),
    "STTRABALHO": VARCHAR(1),
    "STRESTAURA": VARCHAR(1),
    "STFESTA": VARCHAR(1),
    "STOUTROLOC": VARCHAR(1),
    "DS_OUTR_LO": VARCHAR(30),
    "UF_ING": VARCHAR(2),
    "MUN_ING": VARCHAR(6),
    "NUCONSOME": NUMERIC(4),
    "STVENTILA": VARCHAR(1),
    "STANTIBIO": VARCHAR(1),
    "STSORO": VARCHAR(1),
    "STOUTROTRA": VARCHAR(1),
    "DS_TRAT": VARCHAR(30),
    "DTSORO": DATE,
    "STANTIBOTU": VARCHAR(1),
    "STSOROMAT": VARCHAR(1),
    "DTSOROCOL": DATE,
    "STSORORES": VARCHAR(1),
    "TPSOROTOX": VARCHAR(1),
    "STFEZESMAT": VARCHAR(1),
    "DTFEZESCOL": DATE,
    "STFEZESRES": VARCHAR(1),
    "TPFEZESTOX": VARCHAR(1),
    "DS_ALI1OUT": VARCHAR(30),
    "ST_ALI1COL": VARCHAR(1),
    "DT_ALI1COL": DATE,
    "RESALIM1": VARCHAR(1),
    "TP_ALI1TOX": VARCHAR(1),
    "DS_ALI2OUT": VARCHAR(30),
    "ST_ALI2COL": VARCHAR(1),
    "DT_ALI2COL": DATE,
    "RESALIM2": VARCHAR(1),
    "TP_ALI2TO": VARCHAR(1),
    "DS_OUTRO": VARCHAR(30),
    "TP_COLOUT": VARCHAR(1),
    "DT_COLOUT": DATE,
    "RESALIMOUT": VARCHAR(1),
    "TP_TOXOUTR": VARCHAR(1),
    "TP_LIQUOR": VARCHAR(1),
    "DT_LIQUOR": DATE,
    "NU_CELULA": VARCHAR(5),
    "NU_PROTEI": VARCHAR(5),
    "STELETRO": VARCHAR(1),
    "DTELETRO": DATE,
    "TP_SENSITI": VARCHAR(1),
    "TP_MOTORA": VARCHAR(1),
    "TP_REPETE": VARCHAR(1),
    "AGENTE_OUT": VARCHAR(30),
    "TPBOTULISM": VARCHAR(1),
    "STCLINICA": VARCHAR(1),
    "STBROMATO": VARCHAR(1),
    "TPCLINICA": VARCHAR(1),
    "TPBROMATO": VARCHAR(1),
    "DSCAUSALIM": VARCHAR(30),
    "DS_ALI1": VARCHAR(30),
    "DS_ALI2": VARCHAR(30),
    "DS_LOCAL1": VARCHAR(30),
    "DS_LOCAL2": VARCHAR(30),
    "AT_ATIVIDA": VARCHAR(2),
    "AT_LAMINA": VARCHAR(1),
    "AT_SINTOMA": VARCHAR(1),
    "DEXAME": DATE,
    "RESULT": VARCHAR(2),
    "PMM": NUMERIC(8),
    "PCRUZ": VARCHAR(1),
    "DSTRAESQUE": VARCHAR(30),
    "DTRATA": DATE,
    "LOC_INF": VARCHAR(60),
    "NU_LOTE_I": VARCHAR(7),
    "CEFALEIA": VARCHAR(1),
    "ABDOMINAL": VARCHAR(1),
    "MIALGIA": VARCHAR(1),
    "NAUSEA": VARCHAR(1),
    "EXANTEMA": VARCHAR(1),
    "DIARREIA": VARCHAR(1),
    "ICTERICIA": VARCHAR(1),
    "HIPEREMIA": VARCHAR(1),
    "PETEQUIAS": VARCHAR(1),
    "HEMORRAG": VARCHAR(1),
    "LINFADENO": VARCHAR(1),
    "CONVULSAO": VARCHAR(1),
    "NECROSE": VARCHAR(1),
    "PROSTACAO": VARCHAR(1),
    "CHOQUE": VARCHAR(1),
    "COMA": VARCHAR(1),
    "HEMORRAGI": VARCHAR(1),
    "RESPIRATO": VARCHAR(1),
    "OLIGURIA": VARCHAR(1),
    "OUTROS": VARCHAR(1),
    "CARRAPATO": VARCHAR(1),
    "CAPIVARA": VARCHAR(1),
    "CAO_GATO": VARCHAR(1),
    "BOVINO": VARCHAR(1),
    "EQUINOS": VARCHAR(1),
    "OUTROANI": VARCHAR(1),
    "ANIM_ESP": VARCHAR(30),
    "FOI_MATA": VARCHAR(1),
    "COUFHOSP": VARCHAR(2),
    "COMUNHOSP": VARCHAR(6),
    "COUNIHOSP": NUMERIC(7),
    "DIAGNO_LAB": VARCHAR(1),
    "DTS1": DATE,
    "DTS2": DATE,
    "IGM_S1": VARCHAR(1),
    "IGG_S1": VARCHAR(1),
    "IGM_S2": VARCHAR(1),
    "IGG_S2": VARCHAR(1),
    "ISOLAMENTO": VARCHAR(1),
    "AGENTE": VARCHAR(30),
    "HISTOPATO": VARCHAR(1),
    "IMUNOHIST": VARCHAR(1),
    "DIAG_DESCA": VARCHAR(30),
    "ZONA": VARCHAR(1),
    "AMBIENTE": VARCHAR(1),
    "ANT_TIPOCO": VARCHAR(1),
    "ANT_OUTROS": VARCHAR(30),
    "ANT_NOMECO": VARCHAR(70),
    "ANT_ENDECO": VARCHAR(60),
    "ANT_DOS_N": VARCHAR(1),
    "ANT_ULTI_D": DATE,
    "CLI_EDEMAG": VARCHAR(1),
    "CLI_PESCOC": VARCHAR(1),
    "CLI_FEBRE": VARCHAR(1),
    "CLI_PROSTR": VARCHAR(1),
    "CLI_PSEUDO": VARCHAR(1),
    "CLI_PALIDE": VARCHAR(1),
    "CLI_TEMPER": NUMERIC(3),
    "CLI_CAVIDA": VARCHAR(1),
    "CLI_AMIGDA": VARCHAR(1),
    "CLI_CORDAO": VARCHAR(1),
    "CLI_FARING": VARCHAR(1),
    "CLI_LARING": VARCHAR(1),
    "CLI_ORGAOS": VARCHAR(1),
    "CLI_PALATO": VARCHAR(1),
    "CLI_CONDUT": VARCHAR(1),
    "CLI_TRAQUE": VARCHAR(1),
    "CLI_PELE": VARCHAR(1),
    "CLI_CONJUN": VARCHAR(1),
    "CLI_MIOCAR": VARCHAR(1),
    "CLI_NEFRIT": VARCHAR(1),
    "CLI_PARALB": VARCHAR(1),
    "CLI_PARALP": VARCHAR(1),
    "CLI_ARRITM": VARCHAR(1),
    "CLI_PARALM": VARCHAR(1),
    "CLI_OUTRAS": VARCHAR(1),
    "CLI_ESPECI": VARCHAR(30),
    "ATE_HOSPIT": VARCHAR(1),
    "ATE_INTERN": DATE,
    "ATE_UF_INT": VARCHAR(2),
    "ATE_MUNICI": VARCHAR(6),
    "ATE_HOSP_1": NUMERIC((8, 0)),
    "LAB_MATE_N": VARCHAR(1),
    "LAB_DATA_C": DATE,
    "LAB_CULTUR": VARCHAR(1),
    "LAB_PROVAS": VARCHAR(1),
    "TRA_DATA_S": DATE,
    "TRA_ANTIBI": VARCHAR(1),
    "TRA_DATA_A": DATE,
    "MED_IDEN_C": VARCHAR(1),
    "MED_QUAN_C": NUMERIC(3),
    "MED_CASO_S": VARCHAR(1),
    "MED_MATERI": VARCHAR(1),
    "MED_QUAN_M": NUMERIC(3),
    "MED_QUAN_P": NUMERIC(3),
    "MED_PREVEN": VARCHAR(1),
    "VOMITO": VARCHAR(1),
    "DOR_COSTAS": VARCHAR(1),
    "CONJUNTVIT": VARCHAR(1),
    "ARTRITE": VARCHAR(1),
    "ARTRALGIA": VARCHAR(1),
    "PETÉQUIA_N": VARCHAR(1),
    "LEUCOPENIA": VARCHAR(1),
    "LACO": VARCHAR(1),
    "DOR_RETRO": VARCHAR(1),
    "DIABETES": VARCHAR(1),
    "HEMATOLOG": VARCHAR(1),
    "HEPATOPAT": VARCHAR(1),
    "RENAL": VARCHAR(1),
    "HIPERTENSA": VARCHAR(1),
    "ÁCIDO_PEPT": VARCHAR(1),
    "AUTO_IMUNE": VARCHAR(1),
    "DT_CHIK_S1": DATE,
    "DT_CHIK_S2": DATE,
    "DT_PRNT": DATE,
    "RES_CHIKS1": VARCHAR(1),
    "RES_CHIKS2": VARCHAR(1),
    "RESUL_PRNT": VARCHAR(1),
    "DT_SORO": DATE,
    "RESUL_SORO": VARCHAR(1),
    "DT_NS1": DATE,
    "RESUL_NS1": VARCHAR(1),
    "DT_VIRAL": DATE,
    "RESUL_VI_N": VARCHAR(1),
    "DT_PCR": DATE,
    "RESUL_PCR_": VARCHAR(1),
    "SOROTIPO": VARCHAR(1),
    "HISTOPA_N": VARCHAR(1),
    "IMUNOH_N": VARCHAR(1),
    "DDD_HOSP": VARCHAR(2),
    "TEL_HOSP": VARCHAR(9),
    "CLINIC_CHIK": VARCHAR(1),
    "ALRM_HIPOT": VARCHAR(1),
    "ALRM_PLAQ": VARCHAR(1),
    "ALRM_VOM": VARCHAR(1),
    "ALRM_ABDOM": VARCHAR(1),
    "ALRM_LETAR": VARCHAR(1),
    "ALRM_SANG": VARCHAR(1),
    "ALRM_HEMAT": VARCHAR(1),
    "ALRM_HEPAT": VARCHAR(1),
    "ALRM_LIQ": VARCHAR(1),
    "DT_ALRM": DATE,
    "GRAV_PULSO": VARCHAR(1),
    "GRAV_CONV": VARCHAR(1),
    "GRAV_ENCH": VARCHAR(1),
    "GRAV_INSUF": VARCHAR(1),
    "GRAV_TAQUI": VARCHAR(1),
    "GRAV_EXTRE": VARCHAR(1),
    "GRAV_HIPOT": VARCHAR(1),
    "GRAV_HEMAT": VARCHAR(1),
    "GRAV_MELEN": VARCHAR(1),
    "GRAV_METRO": VARCHAR(1),
    "GRAV_SANG": VARCHAR(1),
    "GRAV_AST": VARCHAR(1),
    "GRAV_MIOC": VARCHAR(1),
    "GRAV_CONSC": VARCHAR(1),
    "GRAV_ORGAO": None,
    "DT_GRAV": DATE,
    "ANT_SENTIN": VARCHAR(1),
    "FC_CONTATO": VARCHAR(1),
    "OUT_CONTAT": VARCHAR(30),
    "NM_CONTATO": VARCHAR(40),
    "END_CONTAT": VARCHAR(60),
    "CS_VAC_N": VARCHAR(1),
    "DT_ULT_DOS": DATE,
    "DT_CATARRA": DATE,
    "CS_TOSSE_E": VARCHAR(1),
    "CS_TOSSE_P": VARCHAR(1),
    "CS_CRISE": VARCHAR(1),
    "CS_CIANOSE": VARCHAR(1),
    "CS_VOMITOS": VARCHAR(1),
    "CS_APNEIA": VARCHAR(1),
    "CS_TEMP37": VARCHAR(1),
    "CS_TEMP_38": VARCHAR(1),
    "CS_OUT_SIN": VARCHAR(1),
    "NM_OUT_SIN": VARCHAR(30),
    "CS_PNEUMON": VARCHAR(1),
    "CS_ENCEFAL": VARCHAR(1),
    "CS_DESITRA": VARCHAR(1),
    "CS_OTITE": VARCHAR(1),
    "CS_DESNUTR": VARCHAR(1),
    "CS_OUT_COM": VARCHAR(1),
    "NM_OUT_COM": VARCHAR(30),
    "CS_HOSPITA": VARCHAR(1),
    "COD_UF_HOS": VARCHAR(2),
    "COD_MUN_HO": VARCHAR(6),
    "COD_HOSP": NUMERIC(8),
    "CS_ANTIBIO": VARCHAR(1),
    "DT_ADM_ANT": DATE,
    "CS_COLETA": VARCHAR(1),
    "CS_CULTURA": VARCHAR(1),
    "COLET_COMU": VARCHAR(1),
    "QUAN_COMUN": NUMERIC(3),
    "QUAN_POSIT": NUMERIC(3),
    "MED_BLOQUE": VARCHAR(1),
    "HEPATITE_N": VARCHAR(1),
    "HEPATITA": VARCHAR(1),
    "HEPATITB": VARCHAR(1),
    "INSTITUCIO": VARCHAR(1),
    "HIV": VARCHAR(1),
    "OUTRA_DST": VARCHAR(1),
    "SEXUAL": VARCHAR(1),
    "DOMICILI": VARCHAR(1),
    "OCUPACIO": VARCHAR(1),
    "MEDICAMENT": VARCHAR(1),
    "TATU_PIER": VARCHAR(1),
    "MATBIOLOGI": VARCHAR(1),
    "INAL_CRACK": VARCHAR(1),
    "ACUPUNTURA": VARCHAR(1),
    "TRANSFUSAO": VARCHAR(1),
    "INJETAVEIS": VARCHAR(1),
    "CIRURGICO": VARCHAR(1),
    "AGUA_ALIME": VARCHAR(1),
    "DENTARIO": VARCHAR(1),
    "TRESMAIS": VARCHAR(1),
    "HEMODIALIS": VARCHAR(1),
    "TRANSPLA": VARCHAR(1),
    "OUTRAS": VARCHAR(1),
    "DT_ACIDENT": DATE,
    "CO_UF_EXP": VARCHAR(2),
    "CO_MUN_EXP": VARCHAR(6),
    "DS_LOC_EXP": VARCHAR(70),
    "NU_TEL_EXP": VARCHAR(9),
    "CO_UF_EX2": VARCHAR(2),
    "CO_MUN_EX2": VARCHAR(6),
    "DS_LOC_EX2": VARCHAR(70),
    "NU_TEL_EX2": VARCHAR(9),
    "CO_UF_EX3": VARCHAR(2),
    "CO_MUN_EX3": VARCHAR(6),
    "DS_LOC_EX3": VARCHAR(70),
    "NU_TEL_EX3": VARCHAR(9),
    "BANCOSANGU": VARCHAR(1),
    "RES_HBSAG": VARCHAR(1),
    "RE_ANTIHBC": VARCHAR(1),
    "RE_ANTIHCV": VARCHAR(1),
    "COLETAMARC": DATE,
    "ANTIHAVIGM": VARCHAR(1),
    "ANTIHBS": VARCHAR(1),
    "ANTIHDVIGM": VARCHAR(1),
    "AGHBS": VARCHAR(1),
    "AGHBE": VARCHAR(1),
    "ANTIHEVIGM": VARCHAR(1),
    "ANTIHBCIGM": VARCHAR(1),
    "ANTIHBE": VARCHAR(1),
    "ANTIHCV": VARCHAR(1),
    "HBC_TOTAL": VARCHAR(1),
    "ANTIHDV": VARCHAR(1),
    "TP_SOROHCV": VARCHAR(1),
    "GEN_VHC": VARCHAR(1),
    "FORMA": VARCHAR(1),
    "CLAS_ETIOL": VARCHAR(2),
    "FONTE": VARCHAR(2),
    "DSFONTE": VARCHAR(30),
    "ANT_CB_LAM": VARCHAR(1),
    "ANT_CB_CRI": VARCHAR(1),
    "ANT_CB_CAI": VARCHAR(1),
    "ANT_CB_FOS": VARCHAR(1),
    "ANT_CB_SIN": VARCHAR(1),
    "ANT_CB_PLA": VARCHAR(1),
    "ANT_CB_COR": VARCHAR(1),
    "ANT_CB_ROE": VARCHAR(1),
    "ANT_CB_GRA": VARCHAR(1),
    "ANT_CB_TER": VARCHAR(1),
    "ANT_CB_LIX": VARCHAR(1),
    "ANT_CB_OUT": VARCHAR(1),
    "ANT_OU_DES": VARCHAR(30),
    "ANT_HUMANO": VARCHAR(1),
    "ANT_ANIMAI": VARCHAR(1),
    "CLI_DT_ATE": DATE,
    "CLI_MIALGI": VARCHAR(1),
    "CLI_CEFALE": VARCHAR(1),
    "CLI_PROST": VARCHAR(1),
    "CLI_CONGES": VARCHAR(1),
    "CLI_PANTUR": VARCHAR(1),
    "CLI_VOMITO": VARCHAR(1),
    "CLI_DIARRE": VARCHAR(1),
    "CLI_ICTERI": VARCHAR(1),
    "CLI_RENAL": VARCHAR(1),
    "CLI_RESPIR": VARCHAR(1),
    "CLI_CARDIA": VARCHAR(1),
    "CLI_HEMOPU": VARCHAR(1),
    "CLI_HEMORR": VARCHAR(1),
    "CLI_MENING": VARCHAR(1),
    "CLI_OUTROS": VARCHAR(1),
    "CLI_OTRDES": VARCHAR(30),
    "ATE_HOSP": VARCHAR(1),
    "ATE_DT_INT": DATE,
    "ATE_DT_ALT": DATE,
    "ATE_UF": VARCHAR(2),
    "LAB_DT_1": DATE,
    "LAB_ELIS_1": VARCHAR(1),
    "LAB_DT_2": DATE,
    "LAB_ELIS_2": VARCHAR(1),
    "DTMICRO1": DATE,
    "MICRO1_S1": VARCHAR(5),
    "MICRO1_T_1": VARCHAR(4),
    "MICRO1_S_2": VARCHAR(5),
    "MICRO1_T_2": VARCHAR(5),
    "LAB_MICR_1": VARCHAR(1),
    "DTMICRO2": DATE,
    "MICRO2_S1": VARCHAR(5),
    "MICRO2_T_1": VARCHAR(4),
    "MICRO2_S_2": VARCHAR(5),
    "MICRO2_T_2": VARCHAR(4),
    "LAB_MICR_2": VARCHAR(1),
    "DTISOLA": DATE,
    "RES_ISOL": VARCHAR(1),
    "DTIMUNO": DATE,
    "RES_IMUNO": VARCHAR(1),
    "RES_PCR": VARCHAR(1),
    "CON_AREA": VARCHAR(1),
    "CON_AMBIEN": VARCHAR(1),
    "DT_RISCO1": DATE,
    "DT_RISCO2": DATE,
    "DT_RISCO3": DATE,
    "DT_RISCO4": DATE,
    "CO_MUN_R1": VARCHAR(6),
    "CO_MUN_R2": VARCHAR(6),
    "CO_MUN_R3": VARCHAR(6),
    "CO_MUN_R4": VARCHAR(6),
    "CO_UF_R1": VARCHAR(2),
    "CO_UF_R2": VARCHAR(2),
    "CO_UF_R3": VARCHAR(2),
    "CO_UF_R4": VARCHAR(2),
    "NO_END_R1": VARCHAR(60),
    "NO_END_R2": VARCHAR(60),
    "NO_END_R3": VARCHAR(60),
    "NO_END_R4": VARCHAR(60),
    "NO_LOC_R1": VARCHAR(60),
    "NO_LOC_R2": VARCHAR(60),
    "NO_LOC_R3": VARCHAR(60),
    "NO_LOC_R4": VARCHAR(60),
    "DT_COPRO": DATE,
    "AN_QUANT": NUMERIC(4),
    "AN_QUALI": VARCHAR(1),
    "OUTRO_EX": VARCHAR(40),
    "TRATAM": VARCHAR(1),
    "DTTRAT": DATE,
    "TRATANAO": VARCHAR(1),
    "STCURA1": VARCHAR(1),
    "STCURA2": VARCHAR(1),
    "STCURA3": VARCHAR(1),
    "DT_RESU3": DATE,
    "DS_FORMA": VARCHAR(30),
    "NOPROPIN": VARCHAR(100),
    "NOCOLINF": VARCHAR(100),
    "FRAQUEZA": VARCHAR(1),
    "EMAGRA": VARCHAR(1),
    "TOSSE": VARCHAR(1),
    "PALIDEZ": VARCHAR(1),
    "BACO": VARCHAR(1),
    "INFECCIOSO": VARCHAR(1),
    "FEN_HEMORR": VARCHAR(1),
    "FIGADO": VARCHAR(1),
    "OUTROS_ESP": VARCHAR(30),
    "DIAG_PAR_N": VARCHAR(1),
    "IFI": VARCHAR(1),
    "ENTRADA": VARCHAR(1),
    "TRATAMENTO": VARCHAR(1),
    "PESO": NUMERIC(3),
    "DOSE": VARCHAR(1),
    "AMPOLAS": NUMERIC(3),
    "FALENCIA": VARCHAR(1),
    "DT_DESLC1": DATE,
    "DS_MUN_1": VARCHAR(60),
    "CO_UF_1": VARCHAR(2),
    "CO_PAIS_1": NUMERIC(3),
    "DS_TRANS_1": VARCHAR(30),
    "DT_DESLC2": DATE,
    "DS_MUN_2": VARCHAR(60),
    "CO_UF_2": VARCHAR(2),
    "CO_PAIS_2": NUMERIC(3),
    "DS_TRANS_2": VARCHAR(30),
    "DT_DESLC3": DATE,
    "DS_MUN_3": VARCHAR(60),
    "CO_UF_3": VARCHAR(2),
    "CO_PAIS_3": NUMERIC(3),
    "DS_TRANS_3": VARCHAR(30),
    "TP_CAUSA": VARCHAR(1),
    "TP_CAUSOUT": VARCHAR(30),
    "TP_LOCALLE": VARCHAR(1),
    "NU_DOSE": VARCHAR(1),
    "TP_PROFILA": VARCHAR(1),
    "CS_TRISMO": VARCHAR(1),
    "CS_RISO": VARCHAR(1),
    "CS_OPISTOT": VARCHAR(1),
    "CS_NUCA": VARCHAR(1),
    "CS_ABDOMIN": VARCHAR(1),
    "CS_MEMBROS": VARCHAR(1),
    "CS_CRISES": VARCHAR(1),
    "CS_SIN_OUT": VARCHAR(1),
    "NM_SIN_OUT": VARCHAR(30),
    "TP_ORIGEM": VARCHAR(1),
    "SG_UF_INTE": VARCHAR(2),
    "NM_MUNIC_H": VARCHAR(6),
    "TP_IDENTFI": VARCHAR(1),
    "TP_VACINA": VARCHAR(1),
    "TP_ANALISE": VARCHAR(1),
    "CS_LOCAL": VARCHAR(1),
    "FC_CONT_DE": VARCHAR(30),
    "VINCULO": VARCHAR(1),
    "OUT_VINCUL": VARCHAR(30),
    "CS_ASSINTO": VARCHAR(1),
    "CS_DIARRE": VARCHAR(1),
    "CS_CAIMBRA": VARCHAR(1),
    "CS_FEBRE": VARCHAR(1),
    "CS_DOR": VARCHAR(1),
    "CS_CHOQUE": VARCHAR(1),
    "CS_DESIT": VARCHAR(1),
    "TIP_DIARRE": VARCHAR(1),
    "CS_FREQUEN": VARCHAR(1),
    "CS_SANGUE": VARCHAR(1),
    "CS_MUCO": VARCHAR(1),
    "CS_TIPO": VARCHAR(1),
    "DT_ATENDIM": DATE,
    "UF_HOSPITA": VARCHAR(2),
    "NM_HOSPITA": VARCHAR(70),
    "CS_MATERIA": VARCHAR(1),
    "CS_VOMITO": VARCHAR(1),
    "CS_ANTIB": VARCHAR(1),
    "NM_ANTIBIO": VARCHAR(30),
    "CS_RESULTA": VARCHAR(1),
    "CS_POSITIV": VARCHAR(1),
    "CS_NEG_ESP": VARCHAR(30),
    "CS_REIDRAT": VARCHAR(1),
    "CS_ANTIB_T": VARCHAR(1),
    "ANTIB_DES": VARCHAR(30),
    "NUM_CON_N": VARCHAR(1),
    "CS_VACTETA": VARCHAR(1),
    "DT_1_DOSE": DATE,
    "DT_2_DOSE": DATE,
    "DT_3_DOSE": DATE,
    "DT_REFORCO": DATE,
    "IDADE_MAE": NUMERIC(3),
    "NU_GESTA": VARCHAR(1),
    "ESCOLMAE_N": VARCHAR(2),
    "CS_NASCIDO": VARCHAR(1),
    "NO_OUPARTO": VARCHAR(30),
    "CS_ATEND_N": VARCHAR(1),
    "NO_ATENOUT": VARCHAR(30),
    "CS_SUGOU": VARCHAR(1),
    "CS_MAMAR": VARCHAR(1),
    "CS_CHORO": VARCHAR(1),
    "CS_ABDOMEN": VARCHAR(1),
    "CS_INF_COT": VARCHAR(1),
    "CS_OUTROS": VARCHAR(1),
    "DT_TRISMO": DATE,
    "CS_ORIGEM": VARCHAR(1),
    "CS_COBERTU": VARCHAR(1),
    "NO_COBOUTR": VARCHAR(30),
    "CS_VACINAC": VARCHAR(1),
    "CS_CADASTR": VARCHAR(1),
    "CS_DIVULGA": VARCHAR(1),
    "CS_BUSCAAT": VARCHAR(1),
    "CS_ORIENTA": VARCHAR(1),
    "CS_ANALISE": VARCHAR(1),
    "CS_OUTRAS": VARCHAR(1),
    "NO_OUTRAS": VARCHAR(30),
    "DS_INF_LOC": VARCHAR(1),
    "DS_INF_OUT": VARCHAR(30),
    "COUNIDINF": VARCHAR(7),
    "SIT_TRAB": VARCHAR(2),
    "TRAB_DESC": VARCHAR(30),
    "LOC_EXPO": VARCHAR(1),
    "LOC_EXP_DE": VARCHAR(30),
    "NOEMPRESA": VARCHAR(70),
    "CNAE": VARCHAR(10),
    "UF_EMP": VARCHAR(2),
    "MUN_EMP": VARCHAR(6),
    "DIS_EMP": VARCHAR(9),
    "COBAIEMP": VARCHAR(8),
    "NOBAIEMP": VARCHAR(60),
    "END_EMP": VARCHAR(60),
    "NU_EMP": VARCHAR(6),
    "COMP_EMP": VARCHAR(60),
    "REF_EMP": VARCHAR(60),
    "CEP_EMP": VARCHAR(7),
    "DDD_EMP": VARCHAR(3),
    "FONE_EMP": VARCHAR(9),
    "ZONA_EXP": VARCHAR(1),
    "PAIS_EXP": VARCHAR(4),
    "AGENTE_TOX": VARCHAR(2),
    "OUT_AGENTE": VARCHAR(30),
    "COAGTOXMA1": NUMERIC(3),
    "AGENTE_1": VARCHAR(60),
    "P_ATIVO_1": VARCHAR(60),
    "COAGTOXMA2": NUMERIC(3),
    "AGENTE_2": VARCHAR(60),
    "P_ATIVO_2": VARCHAR(60),
    "COAGTOXMA3": NUMERIC(3),
    "AGENTE_3": VARCHAR(60),
    "P_ATIVO_3": VARCHAR(60),
    "UTILIZACAO": VARCHAR(1),
    "UTIL_DESC": VARCHAR(30),
    "ATIVIDA_1": VARCHAR(2),
    "ATIVIDA_2": VARCHAR(2),
    "ATIVIDA_3": VARCHAR(2),
    "LAVOURA": VARCHAR(100),
    "VIA_1": VARCHAR(1),
    "VIA_2": VARCHAR(1),
    "VIA_3": VARCHAR(1),
    "CIRCUNSTAN": VARCHAR(2),
    "CIRCUN_DES": VARCHAR(30),
    "TPEXP": VARCHAR(1),
    "NUTEMPO": VARCHAR(2),
    "TPTEMPO": VARCHAR(1),
    "TPATENDE": VARCHAR(1),
    "CNES_HOSP": VARCHAR(8),
    "DIAG_CONF": VARCHAR(4),
    "CAT": VARCHAR(1),
    "TREINA_MIL": VARCHAR(1),
    "DESMATA_N": VARCHAR(1),
    "EXPO_N": VARCHAR(1),
    "MOAGEM_N": VARCHAR(1),
    "DORMIU_N": VARCHAR(1),
    "TRANSPO_N": VARCHAR(1),
    "PESCOU_N": VARCHAR(1),
    "ROEDOR_N": VARCHAR(1),
    "OUTRA_ATIV": VARCHAR(1),
    "OUTR_ATI_D": VARCHAR(40),
    "CLI_LOCAL": VARCHAR(30),
    "CLI_TOSSE": VARCHAR(1),
    "CLI_DISPNE": VARCHAR(1),
    "CLI_RESPI": VARCHAR(1),
    "CLI_MIAL_G": VARCHAR(1),
    "CLI_LOMBAR": VARCHAR(1),
    "CLI_ABDOMI": VARCHAR(1),
    "CLI_HIPOTE": VARCHAR(1),
    "CLI_CHOQUE": VARCHAR(1),
    "CLI_TORACI": VARCHAR(1),
    "CLI_TONTUR": VARCHAR(1),
    "CLI_NEUROL": VARCHAR(1),
    "CLI_ASTENI": VARCHAR(1),
    "CLI_PETEQU": VARCHAR(1),
    "CLI_HEMO": VARCHAR(1),
    "CLI_H_DESC": VARCHAR(30),
    "CLI_OUT_D": VARCHAR(30),
    "AM_SANGUE": VARCHAR(1),
    "LAB_HEMA_N": VARCHAR(1),
    "LAB_TROMBO": VARCHAR(1),
    "LAB_ATIPIC": VARCHAR(1),
    "LAB_UREIA": VARCHAR(1),
    "LAB_TGO": VARCHAR(1),
    "LAB_TGO_D": VARCHAR(30),
    "LAB_TGP": VARCHAR(1),
    "LAB_TGP_D": VARCHAR(30),
    "LAB_RES_B": VARCHAR(30),
    "LAB_RADIOL": VARCHAR(1),
    "LAB_DIFUSO": VARCHAR(1),
    "LAB_LOCAL": VARCHAR(1),
    "LAB_DERRAM": VARCHAR(1),
    "DT_COL_IGM": DATE,
    "LAB_IGM_R": VARCHAR(1),
    "LAB_IMUNO": VARCHAR(1),
    "LAB_RTPCR": VARCHAR(1),
    "TRA_HOSP": VARCHAR(1),
    "TRA_DT_INT": DATE,
    "TRA_UF": VARCHAR(2),
    "TRA_MUNICI": VARCHAR(6),
    "TRA_HOSPIT": NUMERIC((8, 0)),
    "TRA_MECANI": VARCHAR(1),
    "TRA_ANTIVI": VARCHAR(1),
    "TRA_CORTIC": VARCHAR(1),
    "TRA_CPAP": VARCHAR(1),
    "TRA_VASOAT": VARCHAR(1),
    "TRA_TRATAM": VARCHAR(1),
    "TRA_ESPECI": VARCHAR(30),
    "CON_FORMA": VARCHAR(1),
    "ZONA_INFEC": VARCHAR(1),
    "CON_AMB_DE": VARCHAR(30),
    "CON_LOCALI": NUMERIC(2),
    "CON_LOCAL2": VARCHAR(1),
    "DT_EVOLUC": DATE,
    "CON_AUTOPS": VARCHAR(1),
    "NU_PRONTUA": VARCHAR(10),
    "POP_LIBER": VARCHAR(1),
    "POP_RUA": VARCHAR(1),
    "POP_SAUDE": VARCHAR(1),
    "POP_IMIG": VARCHAR(1),
    "BENEF_GOV": None,
    "EXTRAPU_N": VARCHAR(2),
    "EXTRAPUL_O": VARCHAR(30),
    "AGRAVAIDS": VARCHAR(1),
    "AGRAVALCOO": VARCHAR(1),
    "AGRAVDIABE": VARCHAR(1),
    "AGRAVDOENC": VARCHAR(1),
    "AGRAVDROGAS": VARCHAR(1),
    "AGRAVTABACO": VARCHAR(1),
    "AGRAVOUTRA": VARCHAR(1),
    "AGRAVOUTDE": VARCHAR(30),
    "BACILOSC_E": VARCHAR(1),
    "RAIOX_TORA": VARCHAR(1),
    "ANTIRRETROVIRAL": VARCHAR(1),
    "HISTOPATOL": VARCHAR(1),
    "CULTURA_ES": VARCHAR(1),
    "TESTE_MOLEC": VARCHAR(1),
    "TEST_SENSIBILID": None,
    "DT_INIC_TR": DATE,
    "NU_COMU_ID": NUMERIC(2),
    "SG_UF_ATUAL": None,
    "ID_MUNIC_AT": None,
    "NU_NOTI_AT": None,
    "DT_NOTI_AT": DATE,
    "ID_UNID_AT": VARCHAR(7),
    "SG_UF_2": None,
    "ID_MUNIC_2": None,
    "NU_CEP2": None,
    "D_DISTR_2": None,
    "ID_BAIRRO2 NM_BAIRRO2": None,
    "BACILOSC_1": None,
    "BACILOSC_2": None,
    "BACILOSC_3": None,
    "BACILOSC_4": None,
    "BACILOSC_5": None,
    "BACILOSC_6": None,
    "BACILOSC_APOS_6": None,
    "NU_PRONT_AT": None,
    "TRATSUP_AT": None,
    "NU_CONT_EX": None,
    "SITUA_ENCE": None,
    "TRANSF": None,
    "SG_UF_TRANSF": None,
    "MUN_TRANSF": None,
    "DT_ENCERRA  ": None,
    "OPORTU": None,
    "DT_OPORTU": None,
    "CONTATO": VARCHAR(1),
    "CONT_OUT": VARCHAR(30),
    "NM_CONTAT": VARCHAR(70),
    "DDD": VARCHAR(2),
    "TEL_CONTAT": VARCHAR(9),
    "SUGE_VINCU": VARCHAR(1),
    "VINC_OUT": VARCHAR(30),
    "ASSINTOMAT": VARCHAR(1),
    "CONSTIPA": VARCHAR(1),
    "ESPLENO": VARCHAR(1),
    "TIFICA": VARCHAR(1),
    "NAUSEAS": VARCHAR(1),
    "VOMITOS": VARCHAR(1),
    "DOR": VARCHAR(1),
    "PULSO": VARCHAR(1),
    "ENTERO": VARCHAR(1),
    "PERFURA": VARCHAR(1),
    "COMP_OUT": VARCHAR(1),
    "COMP_OUT_D": VARCHAR(30),
    "ATENDIMENT": VARCHAR(1),
    "DT_ATENDE": DATE,
    "SANGUE": VARCHAR(1),
    "FEZES": VARCHAR(1),
    "URINA": VARCHAR(1),
    "ANTIBIOTIC": VARCHAR(1),
    "DT_HEMO1": DATE,
    "HEMO_R1": VARCHAR(1),
    "HEMO_D_1": VARCHAR(30),
    "DT_HEMO2": DATE,
    "HEMO_R2": VARCHAR(1),
    "HEMO_D_2": VARCHAR(30),
    "DT_HEMO3": DATE,
    "HEMO_R3": VARCHAR(1),
    "HEMO_D_3": VARCHAR(30),
    "DT_URO": DATE,
    "URO_R1": VARCHAR(1),
    "URO_D": VARCHAR(30),
    "DT_URO2": DATE,
    "URO_R2": VARCHAR(1),
    "URO_D_2": VARCHAR(30),
    "DT_URO3": DATE,
    "URO_R3": VARCHAR(1),
    "URO_D_3": VARCHAR(30),
    "DT_COPRO1": DATE,
    "COPRO_R1": VARCHAR(1),
    "COPRO_D_1": VARCHAR(30),
    "DT_COPRO2": DATE,
    "COPRO_R2": VARCHAR(1),
    "COPRO_D_2": VARCHAR(30),
    "DT_COPRO3": DATE,
    "COPRO_R3": VARCHAR(1),
    "COPRO_D_3": VARCHAR(30),
    "DT_OUTR1": DATE,
    "OUTR_R1": VARCHAR(1),
    "OUTR_D1": VARCHAR(30),
    "DT_OUTR2": DATE,
    "OUTR_R2": VARCHAR(1),
    "OUTR_D2": VARCHAR(30),
    "DT_OUTR3": DATE,
    "OUTR_R3": VARCHAR(1),
    "OUTR_D3": VARCHAR(30),
    "CLORAFEN": VARCHAR(1),
    "AMPICILINA": VARCHAR(1),
    "SULFA": VARCHAR(1),
    "QUINOLONA": VARCHAR(1),
    "ANT_OUTR": VARCHAR(1),
    "ANT_OUT_D": VARCHAR(30),
    "DIAS": VARCHAR(2),
    "ANT_DT_ACI": DATE,
    "ANT_UF": VARCHAR(2),
    "ANT_MUNIC_": VARCHAR(6),
    "ANT_LOCALI": VARCHAR(60),
    "ANT_ZONA": VARCHAR(1),
    "ANT_TEMPO_": VARCHAR(1),
    "ANT_LOCA_1": VARCHAR(1),
    "MCLI_LOCAL": VARCHAR(1),
    "CLI_DOR": VARCHAR(1),
    "CLI_EDEMA": VARCHAR(1),
    "CLI_EQUIMO": VARCHAR(1),
    "CLI_NECROS": VARCHAR(1),
    "CLI_LOCAL_": VARCHAR(1),
    "CLI_LOCA_1": VARCHAR(30),
    "MCLI_SIST": VARCHAR(1),
    "CLI_NEURO": VARCHAR(1),
    "CLI_VAGAIS": VARCHAR(1),
    "CLI_MIOLIT": VARCHAR(1),
    "CLI_OUTR_2": VARCHAR(1),
    "CLI_OUTR_3": VARCHAR(30),
    "CLI_TEMPO_": VARCHAR(1),
    "TP_ACIDENT": VARCHAR(1),
    "ANI_TIPO_1": VARCHAR(30),
    "ANI_SERPEN": VARCHAR(1),
    "ANI_ARANHA": VARCHAR(1),
    "ANI_LAGART": VARCHAR(1),
    "TRA_CLASSI": VARCHAR(1),
    "CON_SOROTE": VARCHAR(1),
    "NU_AMPOLAS": NUMERIC(2),
    "NU_AMPOL_1": NUMERIC(2),
    "NU_AMPOL_8": NUMERIC(2),
    "NU_AMPOL_6": NUMERIC(2),
    "NU_AMPOL_4": NUMERIC(2),
    "NU_AMPOL_7": NUMERIC(2),
    "NU_AMPOL_5": NUMERIC(2),
    "NU_AMPOL_9": NUMERIC(2),
    "NU_AMPOL_3": NUMERIC(2),
    "COM_LOC": VARCHAR(1),
    "COM_SECUND": VARCHAR(1),
    "COM_NECROS": VARCHAR(1),
    "COM_COMPAR": VARCHAR(1),
    "COM_DEFICT": VARCHAR(1),
    "COM_APUTAC": VARCHAR(1),
    "COM_SISTEM": VARCHAR(1),
    "COM_RENAL": VARCHAR(1),
    "COM_EDEMA": VARCHAR(1),
    "COM_SEPTIC": VARCHAR(1),
    "COM_CHOQUE": VARCHAR(1),
    "CLI_CUTANE": VARCHAR(1),
    "CLI_MUCOSA": VARCHAR(1),
    "CLI_CICATR": VARCHAR(1),
    "CLI_CO_HIV": VARCHAR(1),
    "LAB_PARASI": VARCHAR(1),
    "LAB_IRM": VARCHAR(1),
    "LAB_HISTOP": VARCHAR(1),
    "CLA_TIPO_N": VARCHAR(1),
    "CLAS_FORMA": VARCHAR(1),
    "TRA_DROGA_": VARCHAR(1),
    "TRA_PESO": NUMERIC(3),
    "TRA_DOSE": VARCHAR(1),
    "TRA_AMPOLA": NUMERIC(3),
    "TRA_OUTR_N": VARCHAR(1),
    "CON_CLASS_": VARCHAR(1),
    "CO_RISCO": VARCHAR(1),
    "EPI_PESTE": VARCHAR(1),
    "COM_PEST": VARCHAR(1),
    "SIN_GANG": VARCHAR(1),
    "SIN_PULM": VARCHAR(1),
    "TB_INVESTIGA_PESTE": VARCHAR(1),
    "LAB_HEMO": VARCHAR(1),
    "LAB_ESFR": VARCHAR(1),
    "DT_S1": DATE,
    "DT_S2": DATE,
    "ELISA1": VARCHAR(1),
    "ELISA2": VARCHAR(1),
    "HEMO_IGM": VARCHAR(1),
    "IGM_T1": VARCHAR(2),
    "HEMO_IGG": VARCHAR(1),
    "IGG_T2": VARCHAR(5),
    "TRATADO": VARCHAR(1),
    "CO_FOCAL": VARCHAR(1),
    "CON_CLASSI": VARCHAR(1),
    "CON_GRAVID": VARCHAR(1),
    "NU_LESOES": NUMERIC(2),
    "FORMACLINI": VARCHAR(1),
    "CLASSOPERA": VARCHAR(1),
    "NERVOSAFET": NUMERIC(2),
    "AVALIA_N": VARCHAR(1),
    "MODOENTR": VARCHAR(1),
    "MODODETECT": VARCHAR(1),
    "BACILOSCO": VARCHAR(1),
    "DTINICTRAT": DATE,
    "ESQ_INI_N": VARCHAR(1),
    "CONTREG": NUMERIC(2),
    "MIGRADO_W": VARCHAR(1),
    "UFATUAL": VARCHAR(2),
    "ID_MUNI_AT": VARCHAR(6),
    "NU_NOT_AT": VARCHAR(7),
    "UFRESAT": VARCHAR(2),
    "MUNIRESAT": VARCHAR(6),
    "CEP": VARCHAR(8),
    "DISTRIT_AT": VARCHAR(60),
    "BAIRROAT": NUMERIC(8),
    "NOBAIRROAT": VARCHAR(60),
    "DTULTCOMP": DATE,
    "CLASSATUAL": VARCHAR(1),
    "AVAL_ATU_N": VARCHAR(1),
    "ESQ_ATU_N": VARCHAR(1),
    "DOSE_RECEB": NUMERIC(2),
    "EPIS_RACIO": VARCHAR(1),
    "DTMUDESQ": DATE,
    "CONTEXAM": NUMERIC(2),
    "TPALTA_N": VARCHAR(1),
    "DTALTA_N": DATE,
    "IN_VINCULA": VARCHAR(1),
    "NU_LOTE_IA": VARCHAR(7),
    "PRE_MUNIRE": VARCHAR(6),
    "PRE_UNIPRE": NUMERIC((8, 0)),
    "PRE_SISPRE": VARCHAR(10),
    "TPEVIDENCI": VARCHAR(1),
    "TPTESTE1": VARCHAR(1),
    "DSTITULO1": VARCHAR(30),
    "DTTESTE1": DATE,
    "TPCONFIRMA": VARCHAR(10),
    "TPESQUEMA": VARCHAR(1),
    "TRATPARC": VARCHAR(1),
    "TPESQPAR": VARCHAR(1),
    "TPMOTPARC": VARCHAR(1),
    "DSMOTIVO": VARCHAR(30),
    "ARRANHAO": VARCHAR(1),
    "LAMBEDURA": VARCHAR(1),
    "MORDEDURA": VARCHAR(1),
    "MUCOSA": VARCHAR(1),
    "CABECA": VARCHAR(1),
    "MAOS_N": VARCHAR(1),
    "PES": VARCHAR(1),
    "TRONCO": VARCHAR(1),
    "SUPERIORES": VARCHAR(1),
    "INFERIORES": VARCHAR(1),
    "FERIMENT_N": VARCHAR(1),
    "PROFUNDO": VARCHAR(1),
    "SUPERFICIA": VARCHAR(1),
    "DILACERANT": VARCHAR(1),
    "DT_EXPO": DATE,
    "ANTEC_PRE": VARCHAR(1),
    "ANTEC_POS": VARCHAR(1),
    "NUM_DOSES": NUMERIC(2),
    "DT_TR_RAB": DATE,
    "ESPECIE_N": VARCHAR(1),
    "ESP_OUT": VARCHAR(30),
    "VACINAD": VARCHAR(1),
    "AEROFOBIA": VARCHAR(1),
    "HIDROFOBI": VARCHAR(1),
    "DISFAGIA": VARCHAR(1),
    "PARESTESI": VARCHAR(1),
    "AGRESSIVI": VARCHAR(1),
    "PARALISIA": VARCHAR(1),
    "AGITACAO": VARCHAR(1),
    "ANTI_RAB": VARCHAR(1),
    "DT_R_TRA": DATE,
    "DOSES_A": NUMERIC(2),
    "DT_VAC1": DATE,
    "DT_VAC_ULT": DATE,
    "TRA_SORO": VARCHAR(1),
    "DT_APLI_SO": DATE,
    "QUANTID": NUMERIC(3),
    "INFILTRA": VARCHAR(1),
    "IMUNO_DIRE": VARCHAR(1),
    "PROVA_BIOL": VARCHAR(1),
    "IMUNO_INDI": VARCHAR(1),
    "HISTOLOG_N": VARCHAR(1),
    "VARIA_VIR": NUMERIC(2),
    "CON_ZONA": VARCHAR(1),
    "ANT_AC": VARCHAR(1),
    "ANT_DOSE_3": NUMERIC(2),
    "ANT_DTUL_3": DATE,
    "ANT_BC": VARCHAR(1),
    "ANT_DOSES_": NUMERIC(2),
    "ANT_DTULT_": DATE,
    "ANT_CONJ_C": VARCHAR(1),
    "ANT_DOSE_C": NUMERIC(2),
    "ANT_DTUL_C": DATE,
    "ANT_BCG": VARCHAR(1),
    "ANT_DOSE_4": NUMERIC(2),
    "ANT_DTUL_4": DATE,
    "ANT_TRIPLI": VARCHAR(1),
    "ANT_DOSE_5": NUMERIC(2),
    "ANT_DTUL_5": DATE,
    "ANT_HEMO_T": VARCHAR(1),
    "ANT_DOSE_T": NUMERIC(2),
    "ANT_DTUL_T": DATE,
    "ANT_PNEUMO": VARCHAR(1),
    "ANT_DOSE_7": NUMERIC(2),
    "ANT_DTUL_7": DATE,
    "ANT_OUTRA": VARCHAR(1),
    "ANT_OU_DE": VARCHAR(30),
    "ANT_DTUL_8": DATE,
    "ANT_AIDS": VARCHAR(1),
    "ANT_IMUNO": VARCHAR(1),
    "ANT_IRA": VARCHAR(1),
    "ANT_TUBE": VARCHAR(1),
    "ANT_TRAUMA": VARCHAR(1),
    "ANT_INF_HO": VARCHAR(1),
    "ANT_OUTRO": VARCHAR(1),
    "ANT_OUTR_D": VARCHAR(30),
    "A NT_CONT_N": VARCHAR(1),
    "ANT_TELECO": VARCHAR(9),
    "ANT_SECUND": VARCHAR(1),
    "CLI_CONVUL": VARCHAR(1),
    "CLI_RIGIDE": VARCHAR(1),
    "CLI_KERNIG": VARCHAR(1),
    "CLI_ABAULA": VARCHAR(1),
    "CLI_COMA": VARCHAR(1),
    "ATE_UF_HOS": VARCHAR(2),
    "LAB_PUNCAO": VARCHAR(1),
    "LAB_DTPUNC": DATE,
    "LAB_ASPECT": VARCHAR(1),
    "LAB_CTLIQU.": VARCHAR(2),
    "LAB_CTLESA": VARCHAR(2),
    "LAB_CTSANG": VARCHAR(2),
    "LAB_CTESCA": VARCHAR(2),
    "LAB_BCLIQU": VARCHAR(2),
    "LAB_BCLESA": VARCHAR(2),
    "LAB_BCSANG": VARCHAR(2),
    "LAB_BCESCA": VARCHAR(2),
    "LAB_CILIQU": VARCHAR(2),
    "LAB_CISANG": VARCHAR(2),
    "LAB_AGLIQU": VARCHAR(2),
    "LAB_AGSANG": VARCHAR(2),
    "LAB_ISLIQU": VARCHAR(2),
    "LAB_ISFEZE": VARCHAR(2),
    "LAB_PCLIQU": VARCHAR(2),
    "LAB_PCLESA": VARCHAR(2),
    "LAB_PCSANG": VARCHAR(2),
    "LAB_PCESCA": VARCHAR(2),
    "CON_DIAGES": VARCHAR(2),
    "CLA_ME_BAC": VARCHAR(4),
    "CLA_ME_ASS": VARCHAR(4),
    "CLA_ME_ETI": VARCHAR(4),
    "CLA_SOROGR": NUMERIC(4),
    "MED_NUCOMU": NUMERIC(2),
    "MED_QUIMIO": VARCHAR(1),
    "MED_DT_QUI": DATE,
    "MED_DT_EVO": DATE,
    "LAB_HEMA": NUMERIC(5),
    "LAB_NEUTRO": NUMERIC(3),
    "LAB_GLICO": NUMERIC(5),
    "LAB_LEUCO": NUMERIC(5),
    "LAB_EOSI": NUMERIC(3),
    "LAB_PROT": NUMERIC(5),
    "LAB_MONO": NUMERIC(3),
    "LAB_LINFO": NUMERIC(3),
    "LAB_CLOR": NUMERIC(5),
}



# -*- coding:utf-8 -*-
u"""
Created on 2023/12/12
by luabida
license: GPL V3 or Later
"""
import unittest
import datetime

from unittest.mock import patch, MagicMock

from pysus.ftp.databases.pni import PNI
from pysus.ftp import File


class TestPNIDatabase(unittest.TestCase):

    def test_pni(self):
        mock_content = {
            "CPNIAC00.DBF": File(
                path="/dissemin/publicos/PNI/DADOS/CPNIAC00.DBF",
                name="CPNIAC00.DBF",
                info={
                    'size': 14843,
                    'type': 'file',
                    'modify': datetime.datetime(2019, 5, 23, 17, 19)
                }
            ),
            "CPNIAC01.DBF": File(
                path="/dissemin/publicos/PNI/DADOS/CPNIAC01.DBF",
                name="CPNIAC01.DBF",
                info={
                    'size': 14843,
                    'type': 'file',
                    'modify': datetime.datetime(2019, 5, 23, 16, 39)
                }
            ),
            "CPNIAC02.DBF": File(
                path="/dissemin/publicos/PNI/DADOS/CPNIAC02.DBF",
                name="CPNIAC02.DBF",
                info={
                    'size': 14843,
                    'type': 'file',
                    'modify': datetime.datetime(2019, 5, 23, 16, 39)
                }
            ),
        }

        with patch(
            'pysus.ftp.databases.pni.PNI',
            return_value=MagicMock(__content__=mock_content)
        ) as mock_pni:
            pni = PNI()
            pni.__content__ = mock_pni().__content__

            descriptions = [pni.describe(file) for file in pni.files]
            expected_descriptions = [
                {'name': 'CPNIAC00.DBF',
                 'group': 'Cobertura Vacinal',
                 'uf': 'Acre',
                 'year': 2000,
                 'size': '14.8 kB',
                 'last_update': '2019-05-23 05:19PM'},
                {'name': 'CPNIAC01.DBF',
                 'group': 'Cobertura Vacinal',
                 'uf': 'Acre',
                 'year': 2001,
                 'size': '14.8 kB',
                 'last_update': '2019-05-23 04:39PM'},
                {'name': 'CPNIAC02.DBF',
                 'group': 'Cobertura Vacinal',
                 'uf': 'Acre',
                 'year': 2002,
                 'size': '14.8 kB',
                 'last_update': '2019-05-23 04:39PM'}
            ]
            self.assertEqual(descriptions, expected_descriptions)

            formats = [pni.format(file) for file in pni.files]
            expected_formats = [
                ('CPNI', 'AC', 2000),
                ('CPNI', 'AC', 2001),
                ('CPNI', 'AC', 2002)
            ]
            self.assertEqual(formats, expected_formats)

            get_files = pni.get_files(group='CPNI', uf='AC', year=2000)
            self.assertEqual(get_files, [pni.files[0]])



"""
Download SINASC data from DATASUS FTP server
Created on 01/11/17
by fccoelho
license: GPL V3 or Later
"""
from typing import Union

from loguru import logger

from pysus.ftp import CACHEPATH
from pysus.ftp.databases.sinasc import SINASC
from pysus.ftp.utils import parse_UFs

sinasc = SINASC().load()


def get_available_years(group: str, states: Union[str, list[str]]) -> list:
    """
    Get SINASC years for states
    :param group:
        "DN": "Declarações de Nascidos Vivos",
        "DNR": "Dados dos Nascidos Vivos por UF de residência",
    :param states: 2 letter UF code, can be a list. E.g: "SP" or ["SP", "RJ"]
    :return: list of available years
    """
    ufs = parse_UFs(states)

    years = dict()
    for uf in ufs:
        files = sinasc.get_files(group, uf=uf)
        years[uf] = set(sorted([sinasc.describe(f)["year"] for f in files]))

    if len(set([len(v) for v in years.values()])) > 1:
        logger.warning(f"Distinct years were found for UFs: {years}")

    return sorted(list(set.intersection(*map(set, years.values()))))


def download(
    groups: Union[str, list],
    states: Union[str, list],
    years: Union[str, list, int],
    data_dir: str = CACHEPATH,
) -> list:
    """
    Downloads data directly from Datasus ftp server
    :param groups: either DN, DNR or both
    :param states: two-letter state identifier: MG == Minas Gerais,
                   can be a list
    :param years: years to download
    :return: list of downloaded files
    """
    files = sinasc.get_files(groups, uf=states, year=years)
    return sinasc.download(files, local_dir=data_dir)



from typing import List, Union, Optional

from pysus.ftp import Database, Directory, File
from pysus.ftp.utils import zfill_year, to_list, parse_UFs, UFs, MONTHS


class CIHA(Database):
    name = "CIHA"
    paths = (Directory("/dissemin/publicos/CIHA/201101_/Dados"),)
    metadata = {
        "long_name": "Comunicação de Internação Hospitalar e Ambulatorial",
        "source": "http://ciha.datasus.gov.br/CIHA/index.php",
        "description": (
            "A CIHA foi criada para ampliar o processo de planejamento, programação, "
            "controle, avaliação e regulação da assistência à saúde permitindo um "
            "conhecimento mais abrangente e profundo dos perfis nosológico e "
            "epidemiológico da população brasileira, da capacidade instalada e do "
            "potencial de produção de serviços do conjunto de estabelecimentos de saúde "
            "do País. O sistema permite o acompanhamento das ações e serviços de saúde "
            "custeados por: planos privados de assistência à saúde; planos públicos; "
            "pagamento particular por pessoa física; pagamento particular por pessoa "
            "jurídica; programas e projetos federais (PRONON, PRONAS, PROADI); recursos "
            "próprios das secretarias municipais e estaduais de saúde; DPVAT; gratuidade "
            "e, a partir da publicação da Portaria GM/MS nº 2.905/2022, consórcios públicos. "
            "As informações registradas na CIHA servem como base para o processo de "
            "Certificação de Entidades Beneficentes de Assistência Social em Saúde (CEBAS) "
            "e para monitoramento dos programas PRONAS e PRONON."
        ),
    }
    groups = {
        "CIHA": "Comunicação de Internação Hospitalar e Ambulatorial",
    }

    def describe(self, file: File):
        if not isinstance(file, File):
            return file

        if file.extension.upper() in [".DBC", ".DBF"]:
            group, _uf, year, month = self.format(file)

            try:
                uf = UFs[_uf]
            except KeyError:
                uf = _uf

            description = {
                "name": str(file.basename),
                "group": self.groups[group],
                "uf": uf,
                "month": MONTHS[int(month)],
                "year": zfill_year(year),
                "size": file.info["size"],
                "last_update": file.info["modify"],
            }

            return description
        return file

    def format(self, file: File) -> tuple:
        group, _uf = file.name[:4].upper(), file.name[4:6].upper()
        year, month = file.name[-4:-2], file.name[-2:]
        return group, _uf, zfill_year(year), month

    def get_files(
        self,
        uf: Optional[Union[List[str], str]] = None,
        year: Optional[Union[list, str, int]] = None,
        month: Optional[Union[list, str, int]] = None,
        group: Union[List[str], str] = "CIHA",
    ) -> List[File]:
        files = list(filter(
            lambda f: f.extension.upper() in [".DBC", ".DBF"], self.files
        ))

        groups = [gr.upper() for gr in to_list(group)]

        if not all(gr in list(self.groups) for gr in groups):
            raise ValueError(
                "Unknown CIHA Group(s): "
                f"{set(groups).difference(list(self.groups))}"
            )

        files = list(filter(lambda f: self.format(f)[0] in groups, files))

        if uf:
            ufs = parse_UFs(uf)
            files = list(filter(lambda f: self.format(f)[1] in ufs, files))

        if year or str(year) in ["0", "00"]:
            years = [zfill_year(str(m)[-2:]) for m in to_list(year)]
            files = list(filter(lambda f: self.format(f)[2] in years, files))

        if month:
            months = [str(y)[-2:].zfill(2) for y in to_list(month)]
            files = list(filter(lambda f: self.format(f)[3] in months, files))

        return files



from typing import List, Union, Optional

from pysus.ftp import Database, Directory, File
from pysus.ftp.utils import zfill_year, to_list, parse_UFs, UFs, MONTHS


class CNES(Database):
    name = "CNES"
    paths = (Directory("/dissemin/publicos/CNES/200508_/Dados"),)
    metadata = {
        "long_name": "Cadastro Nacional de Estabelecimentos de Saúde",
        "source": "https://cnes.datasus.gov.br/",
        "description": (
            "O Cadastro Nacional de Estabelecimentos de Saúde (CNES) é o sistema "
            "de informação oficial de cadastramento de informações de todos os "
            "estabelecimentos de saúde no país, independentemente de sua natureza "
            "jurídica ou de integrarem o Sistema Único de Saúde (SUS). Trata-se do "
            "cadastro oficial do Ministério da Saúde (MS) no tocante à realidade da "
            "capacidade instalada e mão-de-obra assistencial de saúde no Brasil em "
            "estabelecimentos de saúde públicos ou privados, com convênio SUS ou não."
        ),
    }
    groups = {
        "DC": "Dados Complementares",
        "EE": "Estabelecimento de Ensino",
        "EF": "Estabelecimento Filantrópico",
        "EP": "Equipes",
        "EQ": "Equipamentos",
        "GM": "Gestão e Metas",
        "HB": "Habilitação",
        "IN": "Incentivos",
        "LT": "Leitos",
        "PF": "Profissional",
        "RC": "Regra Contratual",
        "SR": "Serviço Especializado",
        "ST": "Estabelecimentos",
    }
    __loaded__ = set()

    def load(
        self,
        groups: Union[str, List[str]] = None,
    ):
        """
        Loads CNES Groups into content. Will convert the files and directories 
        found within FTP Directories into self.content
        """
        if not self.__content__:
            self.paths[0].load()
            self.__content__ |= self.paths[0].__content__

        if groups:
            groups = to_list(groups)

            if not all(
                group in self.groups for group in [gr.upper() for gr in groups]
            ):
                raise ValueError(
                    "Unknown CNES group(s): "
                    f"{set(groups).difference(self.groups)}"
                )

            for group in groups:
                group = group.upper()
                if group not in self.__loaded__:
                    directory = self.__content__[group]
                    directory.load()
                    self.__content__ |= directory.__content__
                    self.__loaded__.add(directory.name)
        return self

    def describe(self, file: File) -> dict:
        if not isinstance(file, File):
            return {}

        if file.name == "GMufAAmm":
            # Leftover
            return {}

        if file.extension.upper() in [".DBC", ".DBF"]:
            group, _uf, year, month = self.format(file)

            try:
                uf = UFs[_uf]
            except KeyError:
                uf = _uf

            description = {
                "name": str(file.basename),
                "group": self.groups[group],
                "uf": uf,
                "month": MONTHS[int(month)],
                "year": zfill_year(year),
                "size": file.info["size"],
                "last_update": file.info["modify"],
            }

            return description
        return {}

    def format(self, file: File) -> tuple:
        group, _uf = file.name[:2].upper(), file.name[2:4].upper()
        year, month = file.name[-4:-2], file.name[-2:]
        return group, _uf, zfill_year(year), month

    def get_files(
        self,
        group: Union[List[str], str],
        uf: Optional[Union[List[str], str]] = None,
        year: Optional[Union[list, str, int]] = None,
        month: Optional[Union[list, str, int]] = None,
    ) -> List[File]:
        if not group:
            raise ValueError("At least one CNES group is required")

        groups = [gr.upper() for gr in to_list(group)]

        self.load(groups)

        files = list(filter(lambda f: f.name[:2] in groups, self.files))

        if uf:
            ufs = parse_UFs(uf)
            files = list(filter(lambda f: f.name[2:4] in ufs, files))

        if year or str(year) in ["0", "00"]:
            years = [str(m)[-2:].zfill(2) for m in to_list(year)]
            files = list(filter(lambda f: f.name[-4:-2] in years, files))

        if month:
            months = [str(y)[-2:].zfill(2) for y in to_list(month)]
            files = list(filter(lambda f: f.name[-2:] in months, files))

        return files



# -*- coding:utf-8 -*-
u"""
Created on 2023/12/11
by luabida
license: GPL V3 or Later
"""
import unittest
import datetime

from unittest.mock import patch, MagicMock

from pysus.ftp.databases.sia import SIA
from pysus.ftp import File


class TestSIADatabase(unittest.TestCase):

    def test_sia(self):
        mock_content = {
            "ABDF1112.dbc": File(
                path="/dissemin/publicos/SIASUS/200801_/Dados/ABDF1112.dbc",
                name="ABDF1112.dbc",
                info={
                    'size': 2971,
                    'type': 'file',
                    'modify': datetime.datetime(2019, 3, 12, 12, 3)
                }
            ),
            "ABMG1112.dbc": File(
                path="/dissemin/publicos/SIASUS/200801_/Dados/ABMG1112.dbc",
                name="ABMG1112.dbc",
                info={
                    'size': 3183,
                    'type': 'file',
                    'modify': datetime.datetime(2019, 3, 12, 12, 3)
                }
            ),
            "ABOAC1502.dbc": File(
                path="/dissemin/publicos/SIASUS/200801_/Dados/ABOAC1502.dbc",
                name="ABOAC1502.dbc",
                info={
                    'size': 3143,
                    'type': 'file',
                    'modify': datetime.datetime(2016, 9, 12, 8, 45)
                }
            ),
        }

        with patch(
            'pysus.ftp.databases.sia.SIA',
            return_value=MagicMock(__content__=mock_content)
        ) as mock_sia:
            sia = SIA()
            sia.__content__ = mock_sia().__content__

            descriptions = [sia.describe(file) for file in sia.files]
            expected_descriptions = [
                {'name': 'ABDF1112.dbc',
                 'group': 'APAC de Cirurgia Bariátrica',
                 'uf': 'Distrito Federal',
                 'month': 'Dezembro',
                 'year': 2011,
                 'size': '3.0 kB',
                 'last_update': '2019-03-12 12:03PM'},
                {'name': 'ABMG1112.dbc',
                 'group': 'APAC de Cirurgia Bariátrica',
                 'uf': 'Minas Gerais',
                 'month': 'Dezembro',
                 'year': 2011,
                 'size': '3.2 kB',
                 'last_update': '2019-03-12 12:03PM'},
                {'name': 'ABOAC1502.dbc',
                 'group': 'APAC de Acompanhamento Pós Cirurgia Bariátrica',
                 'uf': 'Acre',
                 'month': 'Fevereiro',
                 'year': 2015,
                 'size': '3.1 kB',
                 'last_update': '2016-09-12 08:45AM'}
            ]

            self.assertEqual(descriptions, expected_descriptions)

            formats = [sia.format(file) for file in sia.files]
            expected_formats = [
                ('AB', 'DF', 2011, '12'),
                ('AB', 'MG', 2011, '12'),
                ('ABO', 'AC', 2015, '02')
            ]
            self.assertEqual(formats, expected_formats)

            get_files = sia.get_files(
                group='AB', uf='DF', year=2011, month=12
            )
            self.assertEqual(get_files, [sia.files[0]])



from typing import List, Union, Optional

from pysus.ftp import Database, Directory, File
from pysus.ftp.utils import zfill_year, to_list, parse_UFs, UFs


class SINASC(Database):
    name = "SINASC"
    paths = (
        Directory("/dissemin/publicos/SINASC/NOV/DNRES"),
        Directory("/dissemin/publicos/SINASC/ANT/DNRES"),
    )
    metadata = {
        "long_name": "Sistema de Informações sobre Nascidos Vivos",
        "source": "http://sinasc.saude.gov.br/",
        "description": "",
    }
    groups = {
        "DN": "Declarações de Nascidos Vivos",
        "DNR": "Dados dos Nascidos Vivos por UF de residência",
    }

    def describe(self, file: File) -> dict:
        if file.extension.upper() == ".DBC":
            group, _uf, year = self.format(file)

            try:
                uf = UFs[_uf]
            except KeyError:
                uf = _uf

            description = {
                "name": file.basename,
                "group": self.groups[group],
                "uf": uf,
                "year": year,
                "size": file.info["size"],
                "last_update": file.info["modify"],
            }

            return description
        return {}

    def format(self, file: File) -> tuple:
        if file.name == "DNEX2021":
            pass

        year = zfill_year(file.name[-2:])
        charname = "".join([c for c in file.name if not c.isnumeric()])
        group, _uf = charname[:-2], charname[-2:]
        return group, _uf, zfill_year(year)

    def get_files(
        self,
        group: Union[List[str], str],
        uf: Optional[Union[List[str], str]] = None,
        year: Optional[Union[List, str, int]] = None,
    ) -> List[File]:
        files = self.files

        groups = to_list(group)

        files = list(filter(lambda f: self.format(f)[0] in groups, files))

        if uf:
            if "EX" in to_list(uf):
                # DNEX2021
                if len(to_list(uf)) == 1:
                    return []

                to_list(uf).remove("EX")

            ufs = parse_UFs(uf)
            files = list(filter(lambda f: self.format(f)[1] in ufs, files))

        if year or str(year) in ["0", "00"]:
            years = [zfill_year(str(y)[-2:]) for y in to_list(year)]
            files = list(filter(lambda f: self.format(f)[2] in years, files))

        return files



from typing import List, Union, Optional

from pysus.ftp import Database, Directory, File
from pysus.ftp.utils import zfill_year, to_list, parse_UFs, UFs, MONTHS


class SIA(Database):
    name = "SIA"
    paths = (
        Directory("/dissemin/publicos/SIASUS/199407_200712/Dados"),
        Directory("/dissemin/publicos/SIASUS/200801_/Dados"),
    )
    metadata = {
        "long_name": "Sistema de Informações Ambulatoriais",
        "source": "http://sia.datasus.gov.br/principal/index.php",
        "description": (
            "O Sistema de Informação Ambulatorial (SIA) foi instituído pela Portaria "
            "GM/MS n.º 896 de 29 de junho de 1990. Originalmente, o SIA foi concebido "
            "a partir do projeto SICAPS (Sistema de Informação e Controle Ambulatorial "
            "da Previdência Social), em que os conceitos, os objetivos e as diretrizes "
            "criados para o desenvolvimento do SICAPS foram extremamente importantes e "
            "amplamente utilizados para o desenvolvimento do SIA, tais como: (i) o "
            "acompanhamento das programações físicas e orçamentárias; (ii) o "
            "acompanhamento das ações de saúde produzidas; (iii) a agilização do "
            "pagamento e controle orçamentário e financeiro; e (iv) a formação de "
            "banco de dados para contribuir com a construção do SUS."
        ),
    }
    groups = {
        "AB": "APAC de Cirurgia Bariátrica",
        "ABO": "APAC de Acompanhamento Pós Cirurgia Bariátrica",
        "ACF": "APAC de Confecção de Fístula",
        "AD": "APAC de Laudos Diversos",
        "AM": "APAC de Medicamentos",
        "AMP": "APAC de Acompanhamento Multiprofissional",
        "AN": "APAC de Nefrologia",
        "AQ": "APAC de Quimioterapia",
        "AR": "APAC de Radioterapia",
        "ATD": "APAC de Tratamento Dialítico",
        "BI": "Boletim de Produção Ambulatorial individualizado",
        "IMPBO": "",  # TODO
        "PA": "Produção Ambulatorial",
        "PAM": "",  # TODO
        "PAR": "",  # TODO
        "PAS": "",  # TODO
        "PS": "RAAS Psicossocial",
        "SAD": "RAAS de Atenção Domiciliar",
    }

    def describe(self, file: File) -> dict:
        if file.extension.upper() == ".DBC":
            group, _uf, year, month = self.format(file)

            try:
                uf = UFs[_uf]
            except KeyError:
                uf = _uf

            description = {
                "name": str(file.basename),
                "group": self.groups[group],
                "uf": uf,
                "month": MONTHS[int(month)],
                "year": zfill_year(year),
                "size": file.info["size"],
                "last_update": file.info["modify"],
            }

            return description
        return {}

    def format(self, file: File) -> tuple:
        if file.extension.upper() in [".DBC", ".DBF"]:
            digits = ''.join([d for d in file.name if d.isdigit()])
            chars, _ = file.name.split(digits)
            year, month = digits[:2], digits[2:]
            group, _uf = chars[:-2].upper(), chars[-2:].upper()
            return group, _uf, zfill_year(year), month
        return ()

    def get_files(
        self,
        group: Union[List[str], str],
        uf: Optional[Union[List[str], str]] = None,
        year: Optional[Union[list, str, int]] = None,
        month: Optional[Union[list, str, int]] = None,
    ) -> List[File]:
        files = list(filter(
            lambda f: f.extension.upper() in [".DBC", ".DBF"], self.files
        ))

        groups = [gr.upper() for gr in to_list(group)]

        if not all(gr in list(self.groups) for gr in groups):
            raise ValueError(
                "Unknown SIA Group(s): "
                f"{set(groups).difference(list(self.groups))}"
            )

        files = list(filter(lambda f: self.format(f)[0] in groups, files))

        if uf:
            ufs = parse_UFs(uf)
            files = list(filter(lambda f: self.format(f)[1] in ufs, files))

        if year or str(year) in ["0", "00"]:
            years = [zfill_year(str(m)[-2:]) for m in to_list(year)]
            files = list(filter(lambda f: self.format(f)[2] in years, files))

        if month:
            months = [str(y)[-2:].zfill(2) for y in to_list(month)]
            files = list(filter(lambda f: self.format(f)[3] in months, files))

        return files



from typing import List, Union, Optional, Literal

from pysus.ftp import Database, Directory, File
from pysus.ftp.utils import zfill_year, to_list, parse_UFs, UFs


class PNI(Database):
    name = "PNI"
    paths = (
        Directory("/dissemin/publicos/PNI/DADOS"),
    )
    metadata = {
        "long_name": "Sistema de Informações do Programa Nacional de Imunizações",
        "source": (
            "https://datasus.saude.gov.br/acesso-a-informacao/morbidade-hospitalar-do-sus-sih-sus/",
            "https://datasus.saude.gov.br/acesso-a-informacao/producao-hospitalar-sih-sus/",
        ),
        "description": (
            "O SI-PNI é um sistema desenvolvido para possibilitar aos gestores "
            "envolvidos no Programa Nacional de Imunização, a avaliação dinâmica "
            "do risco quanto à ocorrência de surtos ou epidemias, a partir do "
            "registro dos imunobiológicos aplicados e do quantitativo populacional "
            "vacinado, agregados por faixa etária, período de tempo e área geográfica. "
            "Possibilita também o controle do estoque de imunobiológicos necessário "
            "aos administradores que têm a incumbência de programar sua aquisição e "
            "distribuição. Controla as indicações de aplicação de vacinas de "
            "imunobiológicos especiais e seus eventos adversos, dentro dos Centros "
            "de Referências em imunobiológicos especiais."
        ),
    }
    groups = {
        "CPNI": "Cobertura Vacinal",  # TODO: may be incorrect
        "DPNI": "Doses Aplicadas",  # TODO: may be incorrect
    }

    def describe(self, file: File) -> dict:
        if file.extension.upper() in [".DBC", ".DBF"]:
            group, _uf, year = self.format(file)

            try:
                uf = UFs[_uf]
            except KeyError:
                uf = _uf

            description = {
                "name": file.basename,
                "group": self.groups[group],
                "uf": uf,
                "year": zfill_year(year),
                "size": file.info["size"],
                "last_update": file.info["modify"],
            }

            return description
        return {}

    def format(self, file: File) -> tuple:

        if len(file.name) != 8:
            raise ValueError(f"Can't format {file.name}")

        n = file.name
        group, _uf, year = n[:4], n[4:6], n[-2:]
        return group, _uf, zfill_year(year)

    def get_files(
        self,
        group: Union[list, Literal["CNPI", "DPNI"]],
        uf: Optional[Union[List[str], str]] = None,
        year: Optional[Union[list, str, int]] = None,
    ) -> List[File]:
        files = list(filter(
            lambda f: f.extension.upper() in [".DBC", ".DBF"], self.files
        ))

        groups = [gr.upper() for gr in to_list(group)]

        if not all(gr in list(self.groups) for gr in groups):
            raise ValueError(
                "Unknown PNI Group(s): "
                f"{set(groups).difference(list(self.groups))}"
            )

        files = list(filter(lambda f: self.format(f)[0] in groups, files))

        if uf:
            ufs = parse_UFs(uf)
            files = list(filter(lambda f: self.format(f)[1] in ufs, files))

        if year or str(year) in ["0", "00"]:
            years = [zfill_year(str(m)[-2:]) for m in to_list(year)]
            files = list(filter(lambda f: self.format(f)[2] in years, files))

        return files



from pathlib import Path
from typing import Union
import pandas as pd

from pysus.ftp import CACHEPATH
from pysus.ftp.databases.sinan import SINAN


sinan = SINAN().load()


def list_diseases() -> dict:
    """List available diseases on SINAN"""
    return sinan.diseases


def get_available_years(disease_code: str) -> list:
    """
    Fetch available years for data related to specific disease
    :param disease_code: Disease code. See `SINAN.list_diseases` for valid codes
    :return: A list of DBC files from a specific disease found in the FTP Server.
    """
    files = sinan.get_files(dis_code=disease_code)
    return sorted(list(set(sinan.describe(f)["year"] for f in files)))


def download(
    diseases: Union[str, list],
    years: Union[str, list, int],
    data_path: str = CACHEPATH,
) -> list:
    """
    Downloads SINAN data directly from Datasus ftp server.
    :param disease: Disease code according to `agravos`.
    :param years: 4 digit integer, can be a list of years.
    :param data_path: The directory where the file will be downloaded to.
    :return: list of downloaded files.
    """
    files = sinan.get_files(dis_code=diseases, year=years)
    return sinan.download(files, local_dir=data_path)


def metadata_df(disease_code: str) -> pd.DataFrame:
    metadata_file = (
        Path(__file__).parent.parent
        / "metadata"
        / "SINAN"
        / f"{disease_code}.tar.gz"
    )
    if metadata_file.exists():
        df = pd.read_csv(
            metadata_file,
            compression="gzip",
            header=0,
            sep=",",
            quotechar='"',
            error_bad_lines=False,
        )

        return df.iloc[:, 1:]
    else:
        print(f"No metadata available for {disease_code}")
        return



import os
from pathlib import Path, PurePosixPath, PureWindowsPath
from typing import Dict, List, Union

import pandas as pd
from loguru import logger
from pysus.data import dbc_to_dbf, dbf_to_parquet, parse_dftypes


class ParquetSet:
    """
    A local parquet directory or file
    """

    __path__: Union[PurePosixPath, PureWindowsPath]
    info: Dict

    def __init__(self, path: str, _pbar=None) -> None:
        info = {}
        path = Path(path)

        if path.suffix.lower() not in [".parquet", ".dbc", ".dbf"]:
            raise NotImplementedError(f"Unknown file type: {path.suffix}")

        if path.suffix.lower() == ".dbc":
            path = Path(dbc_to_dbf(path, _pbar=_pbar))

        if path.suffix.lower() == ".dbf":
            path = Path(dbf_to_parquet(path, _pbar=_pbar))

        if path.is_dir():
            info["size"] = sum(
                f.stat().st_size for f in path.glob("**/*") if f.is_file()
            )
        else:
            info["size"] = os.path.getsize(str(path))

        self.__path__ = path
        self.info = info

    def __str__(self):
        return str(self.__path__)

    def __repr__(self):
        return str(self.__path__)

    def __hash__(self):
        return hash(str(self.__path__))

    @property
    def path(self) -> str:
        return str(self.__path__)

    def to_dataframe(self) -> pd.DataFrame:
        """
        Read ParquetSet file(s) into a Pandas DataFrame, concatenating the
        parquets into a single dataframe
        """
        parquets = list(map(str, self.__path__.glob("*.parquet")))
        chunks_list = [
            pd.read_parquet(str(f), engine="fastparquet") for f in parquets
        ]
        _df = pd.concat(chunks_list, ignore_index=True)
        return parse_dftypes(_df)


def parse_data_content(
    path: Union[List[str], str], _pbar=None
) -> Union[ParquetSet, List[ParquetSet]]:
    if isinstance(path, str):
        path = [path]
    else:
        path = list(path)

    content = []
    for _path in path:
        data_path = Path(_path)

        if not data_path.exists():
            continue

        if data_path.suffix.lower() in [".dbc", ".dbf", ".parquet"]:
            content.append(ParquetSet(str(data_path), _pbar=_pbar))
        elif data_path.suffix.lower() == ".zip":
            content.append(str(data_path))
        else:
            continue

    if not content:
        logger.warning("path must be absolute")

    if len(content) == 1:
        return content[0]
    return content


class Data:
    """
    A class parser. Receives an (or a list of) absolute path(s) and returns
    the corresponding ParquetSet instances.
    """

    def __new__(
        cls, path: Union[List[str], str], _pbar=None
    ) -> Union[ParquetSet, List[ParquetSet]]:
        return parse_data_content(path, _pbar=_pbar)



import json
from pathlib import Path
from typing import Union


UFs = {
    "BR": "Brasil",
    "AC": "Acre",
    "AL": "Alagoas",
    "AP": "Amapá",
    "AM": "Amazonas",
    "BA": "Bahia",
    "CE": "Ceará",
    "ES": "Espírito Santo",
    "GO": "Goiás",
    "MA": "Maranhão",
    "MT": "Mato Grosso",
    "MS": "Mato Grosso do Sul",
    "MG": "Minas Gerais",
    "PA": "Pará",
    "PB": "Paraíba",
    "PR": "Paraná",
    "PE": "Pernambuco",
    "PI": "Piauí",
    "RJ": "Rio de Janeiro",
    "RN": "Rio Grande do Norte",
    "RS": "Rio Grande do Sul",
    "RO": "Rondônia",
    "RR": "Roraima",
    "SC": "Santa Catarina",
    "SP": "São Paulo",
    "SE": "Sergipe",
    "TO": "Tocantins",
    "DF": "Distrito Federal",
}

MONTHS = {
    1: "Janeiro",
    2: "Fevereiro",
    3: "Março",
    4: "Abril",
    5: "Maio",
    6: "Junho",
    7: "Julho",
    8: "Agosto",
    9: "Setembro",
    10: "Outubro",
    11: "Novembro",
    12: "Dezembro",
}


def get_city_name_by_geocode(geocode: Union[str, int]):
    """
    Returns the Municipality name from its geocode (IBGE)
    :param geocode: 7 digits city code, according to IBGE format
    :return: City name
    """

    with open(f"{Path(__file__).parent}/municipios.json") as muns:
        _mun_decoded = muns.read().encode().decode("utf-8-sig")
        municipalities = json.loads(_mun_decoded)

    mun_by_geocode = {
        mun["geocodigo"]: mun["municipio"] for mun in municipalities
    }

    return mun_by_geocode[int(geocode)]



"""
Download data from the national immunization program
"""
from typing import Union, Literal

from loguru import logger

from pysus.ftp.databases.pni import PNI
from pysus.ftp import CACHEPATH
from pysus.ftp.utils import parse_UFs


pni = PNI().load()


def get_available_years(group, states):
    """
    Fetch available years for `group` and/or `months`.
    :param group: PNI group, options are "CPNI" or "DPNI"
    :param state: UF code, can be a list. E.g: "SP" or ["SP", "RJ"]
    :return: list of available years
    """
    ufs = parse_UFs(states)

    years = dict()
    for uf in ufs:
        files = pni.get_files(group, uf=uf)
        years[uf] = set(sorted([pni.describe(f)["year"] for f in files]))

    if len(set([len(v) for v in years.values()])) > 1:
        logger.warning(f"Distinct years were found for UFs: {years}")

    return sorted(list(set.intersection(*map(set, years.values()))))


def download(
    group: Union[list, Literal["CNPI", "DPNI"]],
    states: Union[str, list],
    years: Union[str, list, int],
    data_dir: str = CACHEPATH,
) -> list:
    """
    Download imunization records for a given States and years.
    :param group: PNI group, options are "CPNI" or "DPNI"
    :param state: uf two letter code, can be a list. E.g: "SP" or ["SP", "RJ"]
    :param year: year in 4 digits, can be a list. E.g: 1 or [1, 2, 3]
    :param data_dir: directory where data will be downloaded
    :return: list of downloaded ParquetData
    """
    files = pni.get_files(group, uf=states, year=years)
    return pni.download(files, local_dir=data_dir)



from typing import List, Union, Optional

from pysus.ftp import Database, Directory, File
from pysus.ftp.utils import zfill_year, to_list, parse_UFs, UFs


class SIM(Database):
    name = "SIM"
    paths = (
        Directory("/dissemin/publicos/SIM/CID10/DORES"),
        Directory("/dissemin/publicos/SIM/CID9/DORES"),
    )
    metadata = {
        "long_name": "Sistema de Informação sobre Mortalidade",
        "source": "http://sim.saude.gov.br",
        "description": "",
    }
    groups = {"CID10": "DO", "CID9": "DOR"}

    def describe(self, file: File) -> dict:
        group, _uf, year = self.format(file)
        _groups = {v: k for k, v in self.groups.items()}

        try:
            uf = UFs[_uf]
        except KeyError:
            uf = _uf

        description = {
            "name": str(file.basename),
            "uf": uf,
            "year": year,
            "group": _groups[group],
            "size": file.info["size"],
            "last_update": file.info["modify"],
        }

        return description

    def format(self, file: File) -> tuple:
        if "CID9" in str(file.path):
            group, _uf, year = file.name[:-4], file.name[-4:-2], file.name[-2:]
        else:
            group, _uf, year = file.name[:-6], file.name[-6:-4], file.name[-4:]
        return group, _uf, zfill_year(year)

    def get_files(
        self,
        group: Union[list[str], str],
        uf: Optional[Union[list[str], str]] = None,
        year: Optional[Union[list, str, int]] = None,
    ) -> List[File]:
        files = self.files

        groups = [self.groups[g.upper()] for g in to_list(group)]

        files = list(filter(lambda f: self.format(f)[0] in groups, files))

        if uf:
            ufs = parse_UFs(uf)
            files = list(filter(lambda f: self.format(f)[1] in ufs, files))

        if year or str(year) in ["0", "00"]:
            years = [zfill_year(y) for y in to_list(year)]
            files = list(filter(lambda f: self.format(f)[2] in years, files))

        return files



"""
Download data from CIHA and CIH (Old)
Hospital and Ambulatorial information system
http://ciha.datasus.gov.br/CIHA/index.php?area=03

by fccoelho
license: GPL V3 or Later
"""
from typing import Union

from loguru import logger

from pysus.ftp.databases.ciha import CIHA
from pysus.ftp import CACHEPATH
from pysus.ftp.utils import parse_UFs

ciha = CIHA().load()


def get_available_years(
    states: Union[list, str] = None,
) -> dict[str:set[int]]:
    """
    Fetch available years for the `states`.
    :param states: UF code. E.g: "SP" or ["SP", "RJ"]
    :return: list of years in integers
    """
    ufs = parse_UFs(states)

    years = dict()
    for uf in ufs:
        files = ciha.get_files(uf=uf)
        years[uf] = set(sorted([ciha.describe(f)["year"] for f in files]))

    if len(set([len(v) for v in years.values()])) > 1:
        logger.warning(f"Distinct years were found for UFs: {years}")

    return sorted(list(set.intersection(*map(set, years.values()))))


def download(
    states: Union[str, list],
    years: Union[str, list, int],
    months: Union[str, list, int],
    data_dir: str = CACHEPATH,
) -> list:
    """
    Download CIHA records for state, year and month and returns the Parquets
    files as a list of PartquetData
    :param months: 1 to 12, can be a list
    :param states: 2 letter state code,
    :param years: 4 digit integer
    """

    files = ciha.get_files(uf=states, year=years, month=months)
    return ciha.download(files, local_dir=data_dir)



import unittest
import pytest

from pysus.utilities.brasil import get_city_name_by_geocode


class TestGetMunNameByGeocode(unittest.TestCase):

    @pytest.mark.timeout(5)
    def test_get_mun_name_by_geocode(self):
        rio = get_city_name_by_geocode(3304557)
        self.assertEqual(rio, "Rio de Janeiro")

        vale = get_city_name_by_geocode(1101757)
        self.assertEqual(vale, "Vale do Anari")

        santa_helena = get_city_name_by_geocode(5219308)
        self.assertEqual(santa_helena, "Santa Helena de Goiás")


if __name__ == "__main__":
    unittest.main()



from typing import List, Union, Optional

from pysus.ftp import Database, Directory, File
from pysus.ftp.utils import zfill_year, to_list, parse_UFs, UFs, MONTHS


class SIH(Database):
    name = "SIH"
    paths = (
        Directory("/dissemin/publicos/SIHSUS/199201_200712/Dados"),
        Directory("/dissemin/publicos/SIHSUS/200801_/Dados"),
    )
    metadata = {
        "long_name": "Sistema de Informações Hospitalares",
        "source": (
            "https://datasus.saude.gov.br/acesso-a-informacao/morbidade-hospitalar-do-sus-sih-sus/",
            "https://datasus.saude.gov.br/acesso-a-informacao/producao-hospitalar-sih-sus/",
        ),
        "description": (
            "A finalidade do AIH (Sistema SIHSUS) é a de transcrever todos os "
            "atendimentos que provenientes de internações hospitalares que "
            "foram financiadas pelo SUS, e após o processamento, gerarem "
            "relatórios para os gestores que lhes possibilitem fazer os pagamentos "
            "dos estabelecimentos de saúde. Além disso, o nível Federal recebe "
            "mensalmente uma base de dados de todas as internações autorizadas "
            "(aprovadas ou não para pagamento) para que possam ser repassados às "
            "Secretarias de Saúde os valores de Produção de Média e Alta complexidade "
            "além dos valores de CNRAC, FAEC e de Hospitais Universitários – em suas "
            "variadas formas de contrato de gestão."
        ),
    }
    groups = {
        "RD": "AIH Reduzida",
        "RJ": "AIH Rejeitada",
        "ER": "AIH Rejeitada com erro",
        "SP": "Serviços Profissionais",
        "CH": "Cadastro Hospitalar",
        "CM": "",  # TODO
    }

    def describe(self, file: File) -> dict:
        if file.extension.upper() in [".DBC", ".DBF"]:
            group, _uf, year, month = self.format(file)

            try:
                uf = UFs[_uf]
            except KeyError:
                uf = _uf

            description = {
                "name": file.basename,
                "group": self.groups[group],
                "uf": uf,
                "month": MONTHS[int(month)],
                "year": zfill_year(year),
                "size": file.info["size"],
                "last_update": file.info["modify"],
            }

            return description
        return {}

    def format(self, file: File) -> tuple:
        group, _uf = file.name[:2].upper(), file.name[2:4].upper()
        year, month = file.name[-4:-2], file.name[-2:]
        return group, _uf, zfill_year(year), month

    def get_files(
        self,
        group: Union[List[str], str],
        uf: Optional[Union[List[str], str]] = None,
        year: Optional[Union[list, str, int]] = None,
        month: Optional[Union[list, str, int]] = None,
    ) -> List[File]:
        files = list(filter(
            lambda f: f.extension.upper() in [".DBC", ".DBF"], self.files
        ))

        groups = [gr.upper() for gr in to_list(group)]

        if not all(gr in list(self.groups) for gr in groups):
            raise ValueError(
                f"Unknown SIH Group(s): "
                f"{set(groups).difference(list(self.groups))}"
            )

        files = list(filter(lambda f: self.format(f)[0] in groups, files))

        if uf:
            ufs = parse_UFs(uf)
            files = list(filter(lambda f: self.format(f)[1] in ufs, files))

        if year or str(year) in ["0", "00"]:
            years = [zfill_year(str(m)[-2:]) for m in to_list(year)]
            files = list(filter(lambda f: self.format(f)[2] in years, files))

        if month:
            months = [str(y)[-2:].zfill(2) for y in to_list(month)]
            files = list(filter(lambda f: self.format(f)[3] in months, files))

        return files



import unittest
import pytest

from pysus.online_data.ESUS import download


class MyTestCase(unittest.TestCase):
    @pytest.mark.skip(reason="This test takes too long")
    @pytest.mark.timeout(5)
    def test_download(self):
        df = download(uf="se")
        self.assertGreater(len(df), 0)


if __name__ == "__main__":
    unittest.main()



# -*- coding:utf-8 -*-
u"""
Created on 2023/12/12
by luabida
license: GPL V3 or Later
"""
import unittest
import datetime

from unittest.mock import patch, MagicMock

from pysus.ftp.databases.cnes import CNES
from pysus.ftp import File


class TestCNESDatabase(unittest.TestCase):

    def test_cnes(self):
        mock_content = {
            "STAC0508.dbc": File(
                path="/dissemin/publicos/CNES/200508_/Dados/ST/STAC0508.dbc",
                name="STAC0508.dbc",
                info={
                    'size': 18515,
                    'type': 'file',
                    'modify': datetime.datetime(2014, 6, 5, 9, 30)
                }
            ),
            "STAC0509.dbc": File(
                path="/dissemin/publicos/CNES/200508_/Dados/ST/STAC0509.dbc",
                name="STAC0509.dbc",
                info={
                    'size': 18713,
                    'type': 'file',
                    'modify': datetime.datetime(2014, 6, 5, 9, 30)
                }
            ),
            "STAC0510.dbc": File(
                path="/dissemin/publicos/CNES/200508_/Dados/ST/STAC0510.dbc",
                name="STAC0510.dbc",
                info={
                    'size': 17665,
                    'type': 'file',
                    'modify': datetime.datetime(2014, 6, 5, 9, 30)
                }
            ),
        }

        with patch(
            'pysus.ftp.databases.cnes.CNES',
            return_value=MagicMock(__content__=mock_content)
        ) as mock_cnes:
            cnes = CNES().load("ST")
            cnes.__content__ = mock_cnes().__content__

            descriptions = [cnes.describe(file) for file in cnes.files]
            expected_descriptions = [
                {'name': 'STAC0508.dbc',
                 'group': 'Estabelecimentos',
                 'uf': 'Acre',
                 'month': 'Agosto',
                 'year': 2005,
                 'size': '18.5 kB',
                 'last_update': '2014-06-05 09:30AM'},
                {'name': 'STAC0509.dbc',
                 'group': 'Estabelecimentos',
                 'uf': 'Acre',
                 'month': 'Setembro',
                 'year': 2005,
                 'size': '18.7 kB',
                 'last_update': '2014-06-05 09:30AM'},
                {'name': 'STAC0510.dbc',
                 'group': 'Estabelecimentos',
                 'uf': 'Acre',
                 'month': 'Outubro',
                 'year': 2005,
                 'size': '17.7 kB',
                 'last_update': '2014-06-05 09:30AM'}
            ]

            self.assertEqual(descriptions, expected_descriptions)

            formats = [cnes.format(file) for file in cnes.files]
            expected_formats = [
                ('ST', 'AC', 2005, '08'),
                ('ST', 'AC', 2005, '09'),
                ('ST', 'AC', 2005, '10')
            ]
            self.assertEqual(formats, expected_formats)

            get_files = cnes.get_files(group='ST', uf='AC', year=2005, month=8)
            self.assertEqual(get_files, [cnes.files[0]])



import unittest
from unittest.mock import MagicMock, patch
from pysus.ftp.databases import ibge_datasus


class IBGEDATASUSTests(unittest.TestCase):

    @patch('pysus.ftp.databases.ibge_datasus.File')
    def test_describe_zip_file(self, mock_file):
        mock_file.extension.upper.return_value = ".ZIP"
        mock_file.name = "POPTBR12.zip"
        mock_file.basename = "POPTBR12.zip"
        mock_file.info = {"size": 100, "modify": "2022-01-01"}

        ibge = ibge_datasus.IBGEDATASUS()
        result = ibge.describe(mock_file)

        self.assertEqual(result, {
            "name": "POPTBR12.zip",
            "year": 2012,
            "size": 100,
            "last_update": "2022-01-01"
        })

    # @patch('pysus.ftp.databases.ibge_datasus.File')
    # def describe_dbf_file(self, mock_file):
    #     mock_file.extension.upper.return_value = ".DBF"
    #     mock_file.name = "file20.dbf"
    #     mock_file.info = {"size": 100, "modify": "2022-01-01"}
    #
    #     ibge = ibge_datasus.IBGEDATASUS()
    #     result = ibge.describe(mock_file)
    #
    #     self.assertEqual(result, {
    #         "name": "file20",
    #         "year": "2020",
    #         "size": 100,
    #         "last_update": "2022-01-01"
    #     })

    # @patch('pysus.ftp.databases.ibge_datasus.File')
    # def describe_other_file(self, mock_file):
    #     mock_file.extension.upper.return_value = ".TXT"
    #     mock_file.name = "file20.txt"
    #
    #     ibge = ibge_datasus.IBGEDATASUS()
    #     result = ibge.describe(mock_file)
    #
    #     self.assertEqual(result, {})

    @patch('pysus.ftp.databases.ibge_datasus.File')
    def format_file(self, mock_file):
        mock_file.name = "file20.zip"

        ibge = ibge_datasus.IBGEDATASUS()
        result = ibge.format(mock_file)

        self.assertEqual(result, "20.zip")

    @patch('pysus.ftp.databases.ibge_datasus.File')
    @patch('pysus.ftp.databases.ibge_datasus.to_list')
    def test_get_files_with_year(self, mock_to_list, mock_file):
        mock_file.extension.upper.return_value = ".ZIP"
        mock_file.basename = "POPTBR12.zip"
        mock_file.name = "POPTBR12"
        mock_to_list.return_value = ["2012"]

        ibge = ibge_datasus.IBGEDATASUS()
        ibge.__content__ = {"POPTBR12.zip": mock_file}
        result = ibge.get_files(year="2012")

        self.assertEqual(result[0].name, mock_file.name)

    @patch('pysus.ftp.databases.ibge_datasus.File')
    def get_files_without_year(self, mock_file):
        mock_file.extension.upper.return_value = ".ZIP"
        mock_file.name = "file20.zip"

        ibge = ibge_datasus.IBGEDATASUS()
        ibge.files = [mock_file]
        result = ibge.get_files()

        self.assertEqual(result, [mock_file])


if __name__ == '__main__':
    unittest.main()



"""
Download Mortality records from SIM Datasus
Created on 12/12/18
by fccoelho
license: GPL V3 or Later
"""
import os
from ftplib import FTP, error_perm
from typing import Union
import pandas as pd
from dbfread import DBF
from loguru import logger

from pysus.ftp import CACHEPATH
from pysus.ftp.databases.sim import SIM
from pysus.ftp.utils import parse_UFs


sim = SIM().load()


def get_available_years(
    group: str,
    states: Union[str, list] = None,
) -> list:
    """
    Get SIH years for group and/or state and returns a list of years
    :param group: CID9 or CID10
    :param states: 2 letter uf code, can be a list. E.g: "SP" or ["SP", "RJ"]
    :return: list of available years
    """
    ufs = parse_UFs(states)

    years = dict()
    for uf in ufs:
        files = sim.get_files(group, uf=uf)
        years[uf] = set(sorted([sim.describe(f)["year"] for f in files]))

    if len(set([len(v) for v in years.values()])) > 1:
        logger.warning(f"Distinct years were found for UFs: {years}")

    return sorted(list(set.intersection(*map(set, years.values()))))


def download(
    groups: Union[str, list],
    states: Union[str, list],
    years: Union[str, list, int],
    data_dir: str = CACHEPATH,
):
    """
    Downloads data directly from Datasus ftp server
    :param groups: either CID9, CID10 or both
    :param states: two-letter state identifier: MG == Minas Gerais
                   can be a list
    :param years: years to download
    :return: a list of downloaded files
    """
    files = sim.get_files(groups, uf=states, year=years)
    return sim.download(files, local_dir=data_dir)


def get_CID10_chapters_table(cache=True):
    """
    Fetch the CID10 chapters table
    :param cache: If set to True, stores data as parquets.
    :return: Pandas DataFrame
    """
    ftp = FTP("ftp.datasus.gov.br")
    ftp.login()
    logger.debug(
        f"Stablishing connection with ftp.datasus.gov.br.\n{ftp.welcome}"
    )
    ftp.cwd("/dissemin/publicos/SIM/CID10/TABELAS")
    logger.debug(
        "Changing FTP work dir to: /dissemin/publicos/SIM/CID10/TABELAS"
    )

    fname = "CIDCAP10.DBF"
    cachefile = os.path.join(
        CACHEPATH, "SIM_" + fname.split(".")[0] + "_.parquet"
    )

    if os.path.exists(cachefile):
        logger.info(f"Local parquet file found at {cachefile}")
        df = pd.read_parquet(cachefile)

        return df

    try:
        ftp.retrbinary("RETR {}".format(fname), open(fname, "wb").write)

    except error_perm:
        raise Exception("Could not download {}".format(fname))

    dbf = DBF(fname, encoding="iso-8859-1")
    df = pd.DataFrame(list(dbf))

    if cache:
        df.to_parquet(cachefile)
        logger.info(f"Data stored as parquet at {cachefile}")

    os.unlink(fname)
    logger.debug(f"{fname} removed")

    return df


def get_CID10_table(cache=True):
    """
    Fetch the CID10 table
    :param cache: If set to True, stores data as parquets.
    :return: Pandas DataFrame
    """
    ftp = FTP("ftp.datasus.gov.br")
    ftp.login()
    logger.debug(
        f"Stablishing connection with ftp.datasus.gov.br.\n{ftp.welcome}"
    )
    ftp.cwd("/dissemin/publicos/SIM/CID10/TABELAS")
    logger.debug(
        "Changing FTP work dir to: /dissemin/publicos/SIM/CID10/TABELAS"
    )

    fname = "CID10.DBF"
    cachefile = os.path.join(
        CACHEPATH, "SIM_" + fname.split(".")[0] + "_.parquet"
    )

    if os.path.exists(cachefile):
        logger.info(f"Local parquet file found at {cachefile}")
        df = pd.read_parquet(cachefile)

        return df

    try:
        ftp.retrbinary("RETR {}".format(fname), open(fname, "wb").write)

    except error_perm:
        raise Exception("Could not download {}".format(fname))

    dbf = DBF(fname, encoding="iso-8859-1")
    df = pd.DataFrame(list(dbf))

    if cache:
        df.to_parquet(cachefile)
        logger.info(f"Data stored as parquet at {cachefile}")

    os.unlink(fname)
    logger.debug(f"{fname} removed")

    return df


def get_CID9_table(cache=True):
    """
    Fetch the CID9 table
    :param cache: If set to True, stores data as parquets.
    :return: Pandas DataFrame
    """
    ftp = FTP("ftp.datasus.gov.br")
    ftp.login()
    logger.debug(
        f"Stablishing connection with ftp.datasus.gov.br.\n{ftp.welcome}"
    )
    ftp.cwd("/dissemin/publicos/SIM/CID9/TABELAS")
    logger.debug(
        "Changing FTP work dir to: /dissemin/publicos/SIM/CID9/TABELAS"
    )

    fname = "CID9.DBF"
    cachefile = os.path.join(
        CACHEPATH, "SIM_" + fname.split(".")[0] + "_.parquet"
    )

    if os.path.exists(cachefile):
        logger.info(f"Local parquet file found at {cachefile}")
        df = pd.read_parquet(cachefile)

        return df

    try:
        ftp.retrbinary("RETR {}".format(fname), open(fname, "wb").write)

    except error_perm:
        raise Exception("Could not download {}".format(fname))

    dbf = DBF(fname, encoding="iso-8859-1")
    df = pd.DataFrame(list(dbf))

    if cache:
        df.to_parquet(cachefile)
        logger.info(f"Data stored as parquet at {cachefile}")

    os.unlink(fname)
    logger.debug(f"{fname} removed")

    return df


def get_municipios(cache=True):
    """
    Get municipality metadata
    :param cache: If set to True, stores data as parquets.
    :return: Pandas DataFrame
    """
    ftp = FTP("ftp.datasus.gov.br")
    ftp.login()
    logger.debug(
        f"Stablishing connection with ftp.datasus.gov.br.\n{ftp.welcome}"
    )
    ftp.cwd("/dissemin/publicos/SIM/CID10/TABELAS")
    logger.debug(
        "Changing FTP work dir to: /dissemin/publicos/SIM/CID10/TABELAS"
    )

    fname = "CADMUN.DBF"
    cachefile = os.path.join(
        CACHEPATH, "SIM_" + fname.split(".")[0] + "_.parquet"
    )

    if os.path.exists(cachefile):
        logger.info(f"Local parquet file found at {cachefile}")
        df = pd.read_parquet(cachefile)

        return df

    try:
        ftp.retrbinary("RETR {}".format(fname), open(fname, "wb").write)

    except:
        raise Exception("Could not download {}".format(fname))

    dbf = DBF(fname, encoding="iso-8859-1")
    df = pd.DataFrame(list(dbf))

    if cache:
        df.to_parquet(cachefile)
        logger.info(f"Data stored as parquet at {cachefile}")

    os.unlink(fname)
    logger.debug(f"{fname} removed")

    return df


def get_ocupations(cache=True):
    """
    Fetch ocupations table
    :param cache: If set to True, stores data as parquets.
    :return: Pandas DataFrame
    """
    ftp = FTP("ftp.datasus.gov.br")
    ftp.login()
    logger.debug(
        f"Stablishing connection with ftp.datasus.gov.br.\n{ftp.welcome}"
    )
    ftp.cwd("/dissemin/publicos/SIM/CID10/TABELAS")
    logger.debug(
        "Changing FTP work dir to: /dissemin/publicos/SIM/CID10/TABELAS"
    )
    fname = "TABOCUP.DBF"
    cachefile = os.path.join(
        CACHEPATH, "SIM_" + fname.split(".")[0] + "_.parquet"
    )

    if os.path.exists(cachefile):
        logger.info(f"Local parquet file found at {cachefile}")
        df = pd.read_parquet(cachefile)

        return df

    try:
        ftp.retrbinary("RETR {}".format(fname), open(fname, "wb").write)

    except:
        raise Exception("Could not download {}".format(fname))

    dbf = DBF(fname, encoding="iso-8859-1")
    df = pd.DataFrame(list(dbf))

    if cache:
        df.to_parquet(cachefile)
        logger.info(f"Data stored as parquet at {cachefile}")

    os.unlink(fname)
    logger.debug(f"{fname} removed")

    return df



# -*- coding:utf-8 -*-
u"""
Created on 2023/12/11
by luabida
license: GPL V3 or Later
"""
import unittest
from datetime import datetime

from unittest.mock import patch, MagicMock

from pysus.ftp.databases.sim import SIM
from pysus.ftp import File


class TestSIMDatabase(unittest.TestCase):

    def test_sim(self):
        date_format = '%Y-%m-%d %I:%M%p'

        mock_content = {
            "DOAC1996.dbc": File(
                path="/dissemin/publicos/SIM/CID10/DORES/DOAC1996.dbc",
                name="DOAC1996.dbc",
                info={
                    'size': 78054.4,
                    'type': 'DBC file',
                    'modify': datetime.strptime('2020-01-31 02:48PM', date_format)
                }
            ),
            "DOAC1997.dbc": File(
                path="/dissemin/publicos/SIM/CID10/DORES/DOAC1997.dbc",
                name="DOAC1997.dbc",
                info={
                    'size': 79084.8,
                    'type': 'DBC file',
                    'modify': datetime.strptime('2020-01-31 02:48PM', date_format)
                }
            ),
            "DOAC1998.dbc": File(
                path="/dissemin/publicos/SIM/CID10/DORES/DOAC1998.dbc",
                name="DOAC1998.dbc",
                info={
                    'size': 79084.8,
                    'type': 'DBC file',
                    'modify': datetime.strptime('2020-01-31 02:48PM', date_format)
                }
            ),
        }

        with patch(
                'pysus.ftp.databases.sim.SIM',
                return_value=MagicMock(__content__=mock_content)
        ) as mock_sim:
            sim = SIM()
            sim.__content__ = mock_sim().__content__

            descriptions = [sim.describe(file) for file in sim.files]
            expected_descriptions = [
                {'name': 'DOAC1996.dbc',
                 'uf': 'Acre',
                 'year': 1996,
                 'group': 'CID10',
                 'size': '78.1 kB',
                 'last_update': '2020-01-31 02:48PM'},
                {'name': 'DOAC1997.dbc',
                 'uf': 'Acre',
                 'year': 1997,
                 'group': 'CID10',
                 'size': '79.1 kB',
                 'last_update': '2020-01-31 02:48PM'},
                {'name': 'DOAC1998.dbc',
                 'uf': 'Acre',
                 'year': 1998,
                 'group': 'CID10',
                 'size': '79.1 kB',
                 'last_update': '2020-01-31 02:48PM'}
            ]

            self.assertEqual(descriptions, expected_descriptions)

            formats = [sim.format(file) for file in sim.files]
            expected_formats = [
                ('DO', 'AC', 1996), ('DO', 'AC', 1997), ('DO', 'AC', 1998)
            ]
            self.assertEqual(formats, expected_formats)

            get_files = sim.get_files(
                group='CID10', uf='AC', year='1996'
            )
            self.assertEqual(get_files, [sim.files[0]])



import datetime
from typing import Union

from pysus.ftp import to_list
from pysus.utilities.brasil import MONTHS, UFs  # noqa


def zfill_year(year: Union[str, int]) -> int:
    """
    Formats a len(2) year into len(4) with the correct year preffix
    E.g: 20 -> 2020; 99 -> 1999
    """
    year = str(year)[-2:].zfill(2)
    current_year = str(datetime.datetime.now().year)[-2:]
    suffix = "19" if str(year) > current_year else "20"
    return int(suffix + str(year))


def parse_UFs(UF: Union[list[str], str]) -> list:
    """
    Formats states abbreviations into correct format and retuns a list.
    Also checks if there is an incorrect UF in the list.
    E.g: ['SC', 'mt', 'ba'] -> ['SC', 'MT', 'BA']
    """
    ufs = [uf.upper() for uf in to_list(UF)]
    if not all(uf in list(UFs) for uf in ufs):
        raise ValueError(f"Unknown UF(s): {set(ufs).difference(list(UFs))}")
    return ufs



import os
import pandas as pd

from datetime import date
from loguru import logger
from elasticsearch import Elasticsearch, helpers

from pysus.ftp import CACHEPATH


def download(uf, cache=True, checkmemory=True):
    """
    Download ESUS data by UF
    :param uf: rj, mg, etc
    :param cache: if results should be cached on disk
    :return: DataFrame if data fits in memory,
        other an iterator of chunks of size 1000.
    """
    uf = uf.lower()
    user = 'user-public-notificacoes'
    pwd = 'Za4qNXdyQNSa9YaA'
    today = date.today()
    dt = today.strftime('_%d_%m_%Y')
    base = f'desc-esus-notifica-estado-{uf}'  # desc-notificacoes-esusve-
    url = f'https://{user}:{pwd}@elasticsearch-saps.saude.gov.br'
    out = f'ESUS_{uf}_{dt}.parquet'

    cachefile = os.path.join(CACHEPATH, out)
    tempfile = os.path.join(CACHEPATH, f'ESUS_temp_{uf.upper()}.csv.gz')
    if os.path.exists(cachefile):
        logger.info(f'Local parquet file found at {cachefile}')
        df = pd.read_parquet(cachefile)
    elif os.path.exists(tempfile):
        logger.info(f'Local csv file found at {tempfile}')
        df = pd.read_csv(tempfile, chunksize=1000)
    else:
        fname = fetch(base, uf, url)
        size = os.stat(fname).st_size
        if size > 50e6 and checkmemory:
            print(f'Downloaded data is to large:{size / 1e6} MB compressed.')
            print(
                'Only loading the first 1000 rows. If your computer has enough'
                + " memory, set 'checkmemory' to False"
            )
            print(f'The full data is in {fname}')
            df = pd.read_csv(fname, chunksize=1000)
        else:
            df = pd.read_csv(fname, low_memory=False)
            print(f'{df.shape[0]} records downloaded.')
            os.unlink(fname)
            if cache:
                df.to_parquet(cachefile)
                logger.info(f'Data stored as parquet at {cachefile}')

    return df


def fetch(base, uf, url):
    UF = uf.upper()
    print(f'Reading ESUS data for {UF}')
    es = Elasticsearch([url], send_get_body_as='POST')
    body = {'query': {'match_all': {}}}
    results = helpers.scan(es, query=body, index=base)
    # df = pd.DataFrame.from_dict(
    # [document['_source'] for document in results]
    # )

    chunker = chunky_fetch(results, 3000)
    h = 1
    tempfile = os.path.join(CACHEPATH, f'ESUS_temp_{UF}.csv.gz')
    for ch in chunker:
        df = pd.DataFrame.from_dict(ch)
        df.sintomas = df['sintomas'].str.replace(
            ';',
            '',
        )  # remove os  ;
        if h:
            df.to_csv(tempfile)
            h = 0
        else:
            df.to_csv(tempfile, mode='a', header=False)
    # df = pd.read_csv('temp.csv.gz')

    return tempfile


def chunky_fetch(results, chunk_size=3000):
    """Fetches data in chunks to preserve memory"""
    data = []
    i = 0
    for d in results:
        data.append(d['_source'])
        i += 1
        if i == chunk_size:
            yield data
            data = []
            i = 0
    else:
        yield data
